from tala.model.sort import STRING


class Entity(object):
    def __init__(self, identifier, text, start, end):
        self._identifier = identifier
        self._text = text
        self._start = start
        self._end = end

    def generate(self):
        return {
            "start": self._start,
            "end": self._end,
            "value": self._text,
            "entity": self._identifier,
        }

    @property
    def text(self):
        return self._text


class SortalEntity(Entity):
    def __init__(self, sort, text, start, end):
        identifier = STRING if sort == STRING else "sort:%s" % sort
        super(SortalEntity, self).__init__(identifier, text, start, end)


class PropositionalEntity(Entity):
    def __init__(self, predicate, text, start, end):
        identifier = "predicate:%s" % predicate
        super(PropositionalEntity, self).__init__(identifier, text, start, end)



# -*- coding: utf-8 -*-

import codecs
import itertools
from StringIO import StringIO
import sys

from tala.nl import languages
from tala.ddd.ddd_xml_compiler import DddXmlCompiler
from tala.ddd.services.constants import UNDEFINED_SERVICE_ACTION_FAILURE
from tala.ddd.grammar.reader import GrammarReader
from tala.nl.gf import rgl_grammar_entry_types as rgl_types, utils
from tala.nl.gf.auto_generator import LowerCaseGfFileWriter
from tala.nl.gf.grammar_entry_types import Constants, Node
from tala.nl.gf.naming import abstract_gf_filename, natural_language_gf_filename, semantic_gf_filename


class MissingEntry(Exception):
    pass


class GrammarProcessingException(Exception):
    pass


class RglGfFilesGeneratorException(Exception):
    pass


class FunctionNames:
    NP_ACTION = 'NpAction'
    VP_ACTION = 'VpAction'


class Directives:
    FUNCTION = "fun"
    CATEGORY = "cat"
    LINEARIZATION_CATEGORY = "lincat"
    LINEARIZATION = "lin"


DEFAULT_ABSTRACT_GRAMMAR = [
    Node(Constants.ACTION, {"name": "top"}, [Node(rgl_types.NOUN_PHRASE, {}, [Node(rgl_types.NOUN, {"ref": "menu"})])]),
    Node(
        Constants.ACTION, {"name": "up"}, [Node(rgl_types.VERB_PHRASE, {}, [Node(rgl_types.VERB, {"ref": "return"})])]
    ),
]

DEFAULT_LEXICONS = {
    languages.ENGLISH: [
        Node(rgl_types.NOUN, {"id": "menu"}, [Node("singular", {}, ["menu"])]),
        Node(rgl_types.VERB, {"id": "return"}, [Node("infinitive", {}, ["return"])]),
    ],
    languages.FRENCH: [
        Node(rgl_types.NOUN, {"id": "menu"}, [Node("singular", {}, ["menu"])]),
        Node(rgl_types.VERB, {"id": "return"}, [Node("infinitive", {}, ["retourner"])]),
    ],
    languages.DUTCH: [
        Node(rgl_types.NOUN, {"id": "menu"}, [Node("singular", {}, ["menu"])]),
        Node(rgl_types.VERB, {"id": "return"}, [Node("infinitive", {}, ["terugkeren"])]),
    ],
    languages.CHINESE: [
        Node(rgl_types.NOUN, {"id": "menu"}, [Node("singular", {}, [u"菜单"])]),
        Node(rgl_types.VERB, {"id": "return"}, [Node("infinitive", {}, [u"返回"])]),
    ],
}

MAX_NUM_ENTITIES_PER_PARSE = 10


class GfGenerator(object):
    def generate_list(self, strings):
        if len(strings) == 0:
            return '(empty_list)'
        else:
            return '(list %s)' % ' '.join(strings)

    def generate_string_as_utterance(self, string):
        return '(strUtt "%s")' % string.strip()

    def generate_gf_concatenation(self, gf_strings):
        if len(gf_strings) == 1:
            return gf_strings[0]
        else:
            return self._generate_concatentation_tree(gf_strings)

    def _generate_concatentation_tree(self, gf_strings):
        if len(gf_strings) == 1:
            return gf_strings[0]
        else:
            return '(concatUtt %s %s)' % (gf_strings[0], self._generate_concatentation_tree(gf_strings[1:]))

    def get_sort_category(self, sort):
        return 'Sort_%s' % sort

    def get_predicate_category(self, predicate):
        return 'Predicate_%s' % predicate

    def generate_disjunction(self, gf_strings):
        if len(gf_strings) > 1:
            return "(%s)" % "|".join(gf_strings)
        else:
            return gf_strings[0]


class GrammarProcessor(object):
    def get_options_from_node(self, node):
        if node is None:
            return []
        elif node.type == Constants.ONE_OF:
            return self.get_contents_of_items(node.children)
        else:
            return [node]

    def get_contents_of_items(self, nodes):
        return [self.get_content_of_item(node) for node in nodes]

    def get_content_of_item(self, node):
        if node.type == Constants.ITEM and len(node.children) == 1:
            return node.children[0]
        else:
            raise GrammarProcessingException("expected ITEM node with a single child but found %s" % node)

    def has_single_child_node(self, node, child_type):
        return self.get_single_child_node(node, child_type, optional=True) is not None

    def get_single_child_node(self, node, child_type, optional=False):
        child_nodes_of_desired_type = [child for child in node.children if child.type == child_type]
        if len(child_nodes_of_desired_type) == 1:
            return child_nodes_of_desired_type[0]
        elif len(child_nodes_of_desired_type) == 0:
            if optional:
                return None
            else:
                raise GrammarProcessingException("expected a child node of type %s in %s" % (child_type, node))
        else:
            raise GrammarProcessingException(
                "expected exactly 1 child node of type %s in %s, was %s" %
                (child_type, node, child_nodes_of_desired_type)
            )

    def is_individual(self, obj):
        return isinstance(obj, Node) and obj.type == Constants.INDIVIDUAL

    def is_individual_with_predicate(self, obj):
        return isinstance(obj, Node) and obj.type == Constants.INDIVIDUAL and "predicate" in obj.parameters

    def is_background_individual(self, form_part, predicate_name):
        return self.is_individual_with_predicate(form_part
                                                 ) and self.get_individual_predicate(form_part) != predicate_name

    def has_background(self, form, predicate):
        if form is None:
            return False
        else:
            form_parts = self.get_utterance_content(form)
            return any([self.is_background_individual(form_part, predicate.get_name()) for form_part in form_parts])

    def get_individual_sort(self, node):
        return node.parameters["sort"]

    def get_individual_predicate(self, node):
        return node.parameters["predicate"]

    def get_individual_argument(self, node):
        if "sort" in node.parameters:
            sort = node.parameters["sort"]
            return Argument(name=sort, category=self.get_sort_category(sort))
        else:
            raise GrammarProcessingException(
                "expected individual parameters to contain sort but got %s" % node.parameters
            )

    def is_string(self, obj):
        return isinstance(obj, basestring)

    def is_text(self, form_parts):
        return all([self.is_string(form_part) for form_part in form_parts])

    def get_text(self, form_parts):
        return ' '.join(form_parts)

    def get_text_from_utterance_node(self, node):
        options = self.get_utterance_options(node)
        if len(options) == 1 and self.is_text(options[0]):
            return self.get_text(options[0])
        else:
            raise GrammarProcessingException("expected utterance to contain pure text but got\n%s" % node)

    def get_utterance_content(self, node):
        if node.type == rgl_types.UTTERANCE:
            return node.children
        else:
            raise GrammarProcessingException("expected an utterance but got\n%s" % node)

    def get_utterance_options(self, node):
        children = self.get_utterance_content(node)
        if len(children) == 1:
            child = children[0]
            if self.is_string(child):
                return [[child]]
            elif child.type == Constants.ONE_OF:
                return self.get_utterance_options_for_items(child.children)
            else:
                raise GrammarProcessingException(
                    "expected single child of utterance to be string or one-of but got %s" % child
                )
        else:
            self._assert_form_parts_are_string_or_individual(children)
            return [children]

    def _assert_form_parts_are_string_or_individual(self, form_parts):
        for form_part in form_parts:
            if not (self.is_string(form_part) or self.is_individual(form_part)):
                raise GrammarProcessingException("expected a string or an individual but got %s" % form_part)

    def get_utterance_options_for_items(self, items):
        result = []
        for item in items:
            form_parts = self.get_item_content(item)
            result += [form_parts]
        return result

    def get_item_content(self, item):
        if isinstance(item, Node) and item.type == Constants.ITEM:
            self._assert_form_parts_are_string_or_individual(item.children)
            return item.children
        else:
            raise GrammarProcessingException("expected an item node but got\n" % item)

    def get_lexicon_field_from_entry(self, lexicon_entry_nodes, field_name):
        for node in lexicon_entry_nodes:
            if node.type == field_name:
                if len(node.children) == 1:
                    return node.children[0]
                else:
                    raise GrammarProcessingException(
                        "expected lexicon field to contain a single node but found %s" % node.children
                    )
        raise GrammarProcessingException("failed to get lexicon field %r in %s" % (field_name, lexicon_entry_nodes))

    def get_lexicon_entry_from_lexicon(self, lexicon_nodes, type_, id_):
        for node in lexicon_nodes:
            if node.type == type_ and node.parameters["id"] == id_:
                if len(node.children) == 1:
                    return node.children
        raise GrammarProcessingException("failed to get lexicon entry for type=%s id=%s" % (type_, id_))

    def nodes_contain(self, nodes, node_type):
        return any([node for node in nodes if node.type == node_type])


class RglGfFilesGenerator(GfGenerator, GrammarProcessor):
    def __init__(self, ddd, ignore_warnings=False, cwd=None):
        super(RglGfFilesGenerator, self).__init__()
        self._ddd = ddd
        self._ddd_name = ddd.name
        self._ontology = self._ddd.ontology
        self._domain = self._ddd.domain
        self._ignore_warnings = ignore_warnings
        self._sort_categories = self._get_sort_categories()
        self._predicate_categories = self._get_predicate_categories()
        self._rule_count = 1

        self._abstract_gf_content = StringIO()
        self._semantic_gf_content = StringIO()
        self._natural_language_gf_content = StringIO()

    def _get_sort_categories(self):
        return dict([(sort, self.get_sort_category(sort)) for sort in self._ontology.get_sorts()])

    def _get_predicate_categories(self):
        return dict([(predicate, self.get_predicate_category(predicate))
                     for predicate in self._ontology.get_predicates()])

    def generate(self, tdm_language_code):
        self._tdm_language_code = tdm_language_code
        self._gf_language_code = self._tdm_to_gf_language_code(tdm_language_code)
        self._grammar = self._load_and_compile_grammar_entries()
        self._default_grammar = self._create_default_grammar()
        self._lexicon_nodes = self._get_lexicon()
        self._add_header_in_generated_gf_files()
        self._add_content_in_generated_gf_files()
        self._add_footer_in_generated_gf_files()

    def write_to_file(self, _):
        self._write_to_gf_files()

    def clear(self):
        self._clear_io_buffers()

    def _tdm_to_gf_language_code(self, tdm_language_code):
        return tdm_language_code.capitalize()

    def _load_and_compile_grammar_entries(self):
        grammar_source = GrammarReader.read(self._tdm_language_code)
        self._grammar_compiler = DddXmlCompiler()
        return self._grammar_compiler.compile_rgl_grammar(
            grammar_source, self._ontology, self._ddd.service_interface, self._tdm_language_code
        )

    def _get_lexicon(self):
        return self._get_lexicon_from_grammar(self._grammar) + self._get_lexicon_from_grammar(self._default_grammar)

    def _get_lexicon_from_grammar(self, grammar):
        for node in grammar.children:
            if node.type == rgl_types.LEXICON:
                return node.children
        return []

    def _get_lexicon_entry_as_nodes(self, type_, id_):
        return self.get_lexicon_entry_from_lexicon(self._lexicon_nodes, type_, id_)

    def get_noun_singular_form(self, form):
        noun_form = self.get_single_child_node(form, "noun")
        noun_lexicon_entry_as_nodes = self._get_lexicon_entry_as_nodes(rgl_types.NOUN, noun_form.parameters["ref"])
        return self.get_lexicon_field_from_entry(noun_lexicon_entry_as_nodes, "singular")

    def get_verb_infinite_form(self, form):
        verb_form = self.get_single_child_node(form, "verb", optional=False)
        verb_lexicon_entry_as_nodes = self._get_lexicon_entry_as_nodes(rgl_types.VERB, verb_form.parameters["ref"])
        return self.get_lexicon_field_from_entry(verb_lexicon_entry_as_nodes, "infinitive")

    def _clear_io_buffers(self):
        self._abstract_gf_content = StringIO()
        self._semantic_gf_content = StringIO()
        self._natural_language_gf_content = StringIO()

    def _name_of_natural_language_gf_file(self):
        return "build/%s/%s" % (
            self._tdm_language_code, natural_language_gf_filename(self._ddd_name, self._tdm_language_code)
        )

    def _add_header_in_generated_gf_files(self):
        self._abstract_gf_content.write(self._generate_abstract_header())
        self._semantic_gf_content.write(self._generate_semantic_header())
        self._natural_language_gf_content.write(self._generate_natural_language_header())

    def _generate_abstract_header(self):
        header = (
            "--# -coding=utf8\n"
            "abstract %s = TDM, Integers ** {\n" % self._ddd_name + "\n"
            "%s\n\n" % Directives.CATEGORY
        )
        for category in self._get_categories_for_sorts_and_predicates():
            header += "%s;\n" % category
        header += "\n%s\n\n" % Directives.FUNCTION
        return header

    def _get_categories_for_sorts_and_predicates(self):
        return self._sort_categories.values() + self._predicate_categories.values()

    def _generate_semantic_header(self):
        return (
            "--# -coding=utf8\n"
            "concrete %s_sem of %s = TDM_sem, Integers_sem ** open Utils_sem in {\n\n" %
            (self._ddd_name, self._ddd_name) + "\nlin\n\n"
        )

    def _generate_natural_language_header(self):
        header_format = (
            "--# -coding=utf8\n"
            "concrete {ddd_name}_{tdm_language_code} of {ddd_name} =\n"
            "  TDM{gf_language_code} - [sysGreet],\n"
            "  Integers{gf_language_code}\n"
            "** open\n"
            "  Utils{gf_language_code},\n"
            "  TDMInterface{gf_language_code},\n"
            "  Syntax{gf_language_code},\n"
            "  Paradigms{gf_language_code},\n"
            "  Prelude\n"
            "\n"
            "in {{\n"
            "{linearization_category_directive}\n\n"
            "{linearization_categories}\n"
            "{linearization_directive}\n\n"
        )
        return header_format.format(
            ddd_name=self._ddd_name,
            tdm_language_code=self._tdm_language_code,
            gf_language_code=self._gf_language_code,
            linearization_category_directive=Directives.LINEARIZATION_CATEGORY,
            linearization_categories=self._generate_linearization_categories(),
            linearization_directive=Directives.LINEARIZATION
        )

    def _generate_linearization_categories(self):
        sort_lines = ["%s = Sort;\n" % category for category in self._sort_categories.values()]
        predicate_lines = ["%s = Pred;\n" % category for category in self._predicate_categories.values()]
        return "".join(sort_lines + predicate_lines)

    def _add_footer_in_generated_gf_files(self):
        self._abstract_gf_content.write("}\n")
        self._semantic_gf_content.write("}\n")
        self._natural_language_gf_content.write("}\n")

    def _write_to_gf_files(self):
        def write_abstract_file():
            with codecs.open(self._name_of_abstract_gf_file(), "w", encoding="utf-8") as abstract_file:
                self._abstract_gf_content.seek(0)
                for line in self._abstract_gf_content:
                    abstract_file.write(line)

        def write_semantic_file():
            with codecs.open(self._name_of_semantic_gf_file(), "w", encoding="utf-8") as semantic_file:
                self._semantic_gf_content.seek(0)
                for line in self._semantic_gf_content:
                    semantic_file.write(line)

        def write_natural_language_file():
            with LowerCaseGfFileWriter.open(
                self._name_of_natural_language_gf_file(), "w", encoding="utf-8"
            ) as natural_language_file:
                self._natural_language_gf_content.seek(0)
                for line in self._natural_language_gf_content:
                    natural_language_file.write(line)

        write_abstract_file()
        self._abstract_gf_content.close()

        write_semantic_file()
        self._semantic_gf_content.close()

        write_natural_language_file()
        self._natural_language_gf_content.close()

    def _name_of_abstract_gf_file(self):
        return "build/%s/%s" % (self._tdm_language_code, abstract_gf_filename(self._ddd_name))

    def _name_of_semantic_gf_file(self):
        return "build/%s/%s" % (self._tdm_language_code, semantic_gf_filename(self._ddd_name))

    def _add_content_in_generated_gf_files(self):
        self._add_content_based_on_ontology()
        if self._ddd.service_interface:
            self._add_content_based_on_service_interface()

    def _add_content_based_on_ontology(self):
        self._generate_actions()
        self._generate_individuals()
        self._generate_mock_individuals_for_dynamic_sorts()
        self._generate_predicates()

    def _generate_actions(self):
        for action in self._ontology.get_actions():
            self._generate_action(action)

    def _generate_action(self, action):
        try:
            forms = self._get_form_as_options(Node(Constants.ACTION, {"name": action}))
            ActionFunctionGenerator(self, action, forms).generate()
        except MissingEntry:
            self._warn_about_missing_entry(
                """How do speakers talk about the action '%s'? Possible contents of the <action> element:

  <verb-phrase>
  <noun-phrase>
  <one-of>""" % action
            )
        self._generate_potential_request(action)

    def _generate_potential_request(self, action):
        form = self._get_form(Node(rgl_types.REQUEST, {"action": action}), optional=True)
        if form:
            self._generate_requests(action, form)

    def _generate_requests(self, action, form):
        options = self.get_utterance_options(form)
        for option in options:
            RequestFunctionGenerator(self, action, option).generate()

    def _get_form(self, key, optional=False):
        node = self._grammar.get_child(key.type, key.parameters)
        if node is not None:
            return node.get_single_child()
        elif self._has_default_form(key):
            return self._get_default_form(key)
        elif optional:
            return None
        else:
            raise MissingEntry("missing entry %s" % unicode(key))

    def _get_form_as_options(self, key, optional=False):
        node = self._get_form(key, optional)
        return self.get_options_from_node(node)

    def _warn_about_missing_entry(self, warning):
        if not self._ignore_warnings:
            print >> sys.stderr, "\nMissing grammar entry:",
            print >> sys.stderr, warning

    def _add_content_based_on_service_interface(self):
        for action_interface in self._ddd.service_interface.actions:
            self._generate_report_ended_and_failed_content(action_interface)
            self._generate_report_started_content(action_interface)
        for validator_interface in self._ddd.service_interface.validators:
            self._generate_validity_content(validator_interface)

    def _generate_report_ended_and_failed_content(self, action_interface):
        key = Node(Constants.REPORT_ENDED, {"action": action_interface.name})
        try:
            form = self._get_form(key)
            ReportEndedFunctionGenerator(self, action_interface.name, form).generate()
            FunctionGeneratorForReportFailedWithUnknownFailure(self, action_interface.name, form).generate()
        except MissingEntry:
            if self._service_action_has_postconfirm(action_interface.name):
                warning = self._get_warning_for_missing_report_ended(key, action_interface.name)
                self._warn_about_missing_entry(warning)

    def _get_warning_for_missing_report_ended(self, key, action_name):
        example_phrase = "performed %s" % action_name
        example_form = Node(rgl_types.UTTERANCE, {}, [example_phrase])
        example_xml = self._decompile_entry(key, example_form)
        return "How does the system report that the service action '%s' ended? Example:\n\n%s\n" % (
            action_name, example_xml
        )

    def _generate_report_started_content(self, action_interface):
        key = Node(Constants.REPORT_STARTED, {"action": action_interface.name})
        try:
            form = self._get_form(key)
            ReportStartedFunctionGenerator(self, action_interface.name, form).generate()
        except MissingEntry:
            pass

    def _generate_predicates(self):
        for predicate in self._ontology.get_predicates().values():
            predicate_name = predicate.get_name()
            self._generate_predicate_content(predicate_name)
            self._generate_potential_system_question(predicate_name)
            self._generate_potential_user_question(predicate_name)
            self._generate_system_answer_content(predicate)
            self._generate_user_answer_content(predicate)

    def _generate_predicate_content(self, predicate_name):
        function_name = predicate_name
        self._abstract_gf_content.write('%s : Predicate;\n' % function_name)
        self._semantic_gf_content.write('%s = pp "%s";\n' % (function_name, predicate_name))
        form = self._get_form(Node(rgl_types.PREDICATE, {"name": predicate_name}), optional=True)
        if form:
            self._generate_speaker_independent_predicate_content(function_name, form)
            self._generate_resolve_ynq(predicate_name, form)

    def _generate_speaker_independent_predicate_content(self, function_name, form):
        noun = self.get_noun_singular_form(form)
        self._natural_language_gf_content.write('%s = mkPred (mkNP the_Det (mkN "%s"));\n' % (function_name, noun))

    def _generate_resolve_ynq(self, predicate_name, form):
        function_name = '%s_resolve_ynq' % predicate_name
        self._abstract_gf_content.write('%s : SysResolveGoal;\n' % function_name)
        self._semantic_gf_content.write('%s = resolve_ynq %s;\n' % (function_name, predicate_name))
        noun = self.get_noun_singular_form(form)
        self._natural_language_gf_content.write(
            '%s = mkSysResolveGoal (mkVP know_V2 (mkNP the_Det (mkN "%s")));\n' % (function_name, noun)
        )

    def _generate_potential_system_question(self, predicate):
        if self._can_be_asked_by_system(predicate):
            self._generate_system_question(predicate)

    def _generate_system_question(self, predicate):
        key = Node(Constants.SYS_QUESTION, {"predicate": predicate})
        try:
            form = self._get_form(key)
            function_name = predicate
            natural_language_content = self._generate_natural_language_for_system_question(form)
            self._natural_language_gf_content.write('%s = %s;\n' % (function_name, natural_language_content))
        except MissingEntry:
            warning = self._get_warning_for_missing_system_question(key, predicate)
            self._warn_about_missing_entry(warning)

    def _can_be_asked_by_system(self, predicate_name):
        for question in self._domain.get_plan_questions():
            if self._question_has_predicate_name(question, predicate_name) and not self._question_is_feature(question):
                return True
            elif self._question_has_ask_feature(question, predicate_name):
                return True

    def _question_has_predicate_name(self, question, predicate_name):
        return question.get_content().is_lambda_abstracted_predicate_proposition() and question.get_predicate(
        ).get_name() == predicate_name

    def _question_is_feature(self, question):
        return question.get_predicate().get_feature_of_name() is not None

    def _question_has_ask_feature(self, question, feature_name):
        ask_features = self._domain.get_ask_features(question)
        if ask_features is not None:
            for feature_predicate in ask_features:
                if feature_predicate.get_name() == feature_name:
                    return True

    def _get_warning_for_missing_system_question(self, key, predicate):
        example_phrase = "what is %s" % self._example_phrase_from_semantic_value(predicate)
        example_form = Node(rgl_types.UTTERANCE, {}, [example_phrase])
        example_xml = self._decompile_entry(key, example_form)
        return "How does the system ask about '%s'?\n\nExample:\n\n%s\n" % (predicate, example_xml)

    def _example_phrase_from_semantic_value(self, string):
        return string.replace("_", " ")

    def _decompile_entry(self, key, value):
        node = Node(key.type, key.parameters, [value])
        return self._decompile_node(node)

    def _decompile_node(self, node):
        return self._grammar_compiler.decompile_grammar_node(self._ddd, self._ddd.language_codes, node)

    def _generate_potential_user_question(self, predicate_name):
        form = self._get_form(Node(Constants.USER_QUESTION, {"predicate": predicate_name}), optional=True)
        if form:
            form_variants = self.get_utterance_options(form)
            for form_variant in form_variants:
                UserQuestionFunctionGenerator(self, predicate_name, form_variant).generate()

    def _generate_natural_language_for_system_question(self, form):
        return 'mkPred "%s"' % self.get_text_from_utterance_node(form)

    def _generate_system_answer_content(self, predicate):
        sort = predicate.getSort()
        if sort.is_boolean_sort():
            raise RglGfFilesGeneratorException("system answers of boolean sort not supported")
        else:
            self._generate_unary_system_answer_content(predicate)

    def _generate_unary_system_answer_content(self, predicate):
        form = self._get_form(Node(Constants.SYS_ANSWER, {"predicate": predicate.get_name()}), optional=True)
        sort = predicate.getSort()
        if sort.is_builtin():
            raise RglGfFilesGeneratorException("system answers of built-in sorts such as %s are not supported" % sort)
        else:
            self._generate_system_answer_of_custom_sort(predicate, form)

    def _generate_system_answer_of_custom_sort(self, predicate, form):
        if self.has_background(form, predicate):
            FunctionGeneratorForSystemAnswerOfCustomSortWithBackground(self, predicate, form).generate()
        else:
            FunctionGeneratorForSystemAnswerOfCustomSortWithoutBackground(self, predicate, form).generate()

    def _generate_user_answer_content(self, predicate):
        sort = predicate.getSort()
        self._generate_sortal_user_answer(predicate, sort)

    def _generate_sortal_user_answer(self, predicate, sort):
        self.write_function(
            function_name='%s_sortal_usr_answer' % predicate,
            category='UsrAnswer',
            arguments=[Argument(name="answer", category=self.get_sort_category(sort.get_name()))],
            semantic_content='answer',
            natural_language_content='mkUsr answer'
        )

    def _generate_individuals(self):
        for individual, sort in self._ontology.get_individuals().iteritems():
            self._generate_individual(individual, sort)

    def _generate_individual(self, individual, sort):
        try:
            key = Node(Constants.INDIVIDUAL, {"name": individual})
            form = self._get_form(key)
        except MissingEntry:
            return
        self.write_function(
            function_name=individual,
            category=self.get_sort_category(sort.get_name()),
            semantic_content='pp "%s"' % individual,
            natural_language_content=self._generate_natural_language_for_individual(form)
        )

    def _generate_natural_language_for_individual(self, form):
        if form.type == rgl_types.PROPER_NOUN and len(form.children) == 1:
            return 'mkSort (mkPN ("%s"))' % form.children[0]
        else:
            raise RglGfFilesGeneratorException("expected proper noun with single child but got %s" % form)

    def _generate_validity_content(self, validator_interface):
        key = Node(Constants.VALIDITY, {"name": validator_interface.name})
        try:
            forms = self._get_form_as_options(key)
            ValidityGenerator(self, validator_interface, forms).generate()
        except MissingEntry:
            warning = 'How does the system report that the device validity %s is unsatisfied?' % validator_interface.name
            self._warn_about_missing_entry(warning)

    def _create_default_grammar(self):
        lexicon = DEFAULT_LEXICONS[self._tdm_language_code]
        return Node(Constants.GRAMMAR, {}, DEFAULT_ABSTRACT_GRAMMAR + [Node(rgl_types.LEXICON, {}, lexicon)])

    def _has_default_form(self, key):
        node = self._default_grammar.get_child(key.type, key.parameters)
        return node is not None

    def _get_default_form(self, key):
        node = self._default_grammar.get_child(key.type, key.parameters)
        if node is not None:
            return node.get_single_child()

    def _generate_mock_individuals_for_dynamic_sorts(self):
        for sort in self._ontology.get_sorts().values():
            if sort.is_dynamic():
                for i in range(0, MAX_NUM_ENTITIES_PER_PARSE):
                    placeholder_name = utils.name_of_user_answer_placeholder_of_sort(sort.get_name(), i)
                    sem_placeholder = utils.semantic_user_answer_placeholder_of_sort(sort.get_name(), i)
                    nl_placeholder = utils.nl_user_answer_placeholder_of_sort(sort.get_name(), i)

                    self._abstract_gf_content.write(
                        "%s : %s;\n" % (placeholder_name, self.get_sort_category(sort.get_name()))
                    )
                    self._semantic_gf_content.write('%s = pp "%s";\n' % (placeholder_name, sem_placeholder))
                    self._natural_language_gf_content.write(
                        '%s = mkSort (mkPN ("%s"));\n' % (placeholder_name, nl_placeholder)
                    )

    def generate_new_function_name(self, prefix):
        function_name = "%s_%d" % (prefix, self._rule_count)
        self._rule_count += 1
        return function_name

    def write_function(self, function_name, category, semantic_content, natural_language_content, arguments=[]):
        self._write_abstract(function_name, category, arguments)
        self._write_semantic(function_name, arguments, semantic_content)
        self._write_natural_language(function_name, arguments, natural_language_content)

    def _write_abstract(self, function_name, category, arguments):
        self._abstract_gf_content.write("%s : " % function_name)
        for argument in arguments:
            self._abstract_gf_content.write("%s -> " % argument.category)
        self._abstract_gf_content.write("%s;\n" % category)

    def _write_semantic(self, function_name, arguments, semantic):
        self._semantic_gf_content.write("%s " % function_name)
        for argument in arguments:
            self._semantic_gf_content.write("%s " % argument.name)
        self._semantic_gf_content.write("= %s;\n" % semantic)

    def _write_natural_language(self, function_name, arguments, natural_language):
        self._natural_language_gf_content.write("%s " % function_name)
        for argument in arguments:
            self._natural_language_gf_content.write("%s " % argument.name)
        self._natural_language_gf_content.write("= %s;\n" % natural_language)

    def _service_action_has_postconfirm(self, action_name):
        for item in self._domain.get_invoke_service_action_items_for_action(action_name):
            if item.has_postconfirmation():
                return True


class FunctionGenerator(GrammarProcessor, GfGenerator):
    def __init__(self, files_generator):
        self._files_generator = files_generator

    def generate(self):
        self._files_generator.write_function(
            function_name=self.get_function_name(),
            category=self.get_function_category(),
            arguments=self.get_arguments(),
            semantic_content=self.get_semantic_content(),
            natural_language_content=self.get_natural_language_content()
        )

    def generate_new_function_name(self, prefix):
        return self._files_generator.generate_new_function_name(prefix)

    def get_arguments(self):
        return []


class ActionFunctionGenerator(FunctionGenerator):
    def __init__(self, files_generator, action, forms):
        self._files_generator = files_generator
        self._action = action
        self._forms = forms

    def get_function_name(self):
        return self._action

    def get_function_category(self):
        forms_contain_noun_phrase = self.nodes_contain(self._forms, rgl_types.NOUN_PHRASE)
        forms_contain_verb_phrase = self.nodes_contain(self._forms, rgl_types.VERB_PHRASE)
        if forms_contain_noun_phrase and forms_contain_verb_phrase:
            raise RglGfFilesGeneratorException(
                "Exception when generating action %s: Cannot mix noun and verb phrase." % self._action
            )
        elif forms_contain_noun_phrase:
            return FunctionNames.NP_ACTION
        elif forms_contain_verb_phrase:
            return FunctionNames.VP_ACTION
        else:
            raise RglGfFilesGeneratorException(
                "Exception when generating action %s: Expected entry to contain either "
                "a noun phrase or a verb phrase." % self._action
            )

    def get_semantic_content(self):
        return 'pp "%s"' % self._action

    def get_natural_language_content(self):
        generated_forms = [self._generate_natural_language_content(form) for form in self._forms]
        return self.generate_disjunction(generated_forms)

    def _generate_natural_language_content(self, form):
        if form.type == rgl_types.NOUN_PHRASE:
            return self._generate_noun_phrase(form)
        elif form.type == rgl_types.VERB_PHRASE:
            return self._generate_verb_phrase(form)
        else:
            raise GrammarProcessingException("don't know how to generate natural language content for %r" % form)

    def _generate_noun_phrase(self, form):
        noun_singular_form = self._files_generator.get_noun_singular_form(form)
        return 'mkNpAction (mkNP (mkPN "%s") | mkNP the_Det (mkCN (mkN "%s")))' % (
            noun_singular_form, noun_singular_form
        )

    def _generate_verb_phrase(self, form):
        verb_infinitive_form = self._files_generator.get_verb_infinite_form(form)
        if self._has_noun(form):
            noun_singular_form = self._files_generator.get_noun_singular_form(form)
            return 'mkVpAction (mkVP (mkV2 (mkV "%s")) (mkNP (a_Det|the_Det) (mkCN (mkN "%s"))))' % (
                verb_infinitive_form, noun_singular_form
            )
        else:
            return 'mkVpAction "%s"' % verb_infinitive_form

    def _has_noun(self, form):
        return self.has_single_child_node(form, "noun")


class UserQuestionFunctionGenerator(FunctionGenerator):
    def __init__(self, rgl_gf_generator, predicate_name, form_parts):
        self._predicate_name = predicate_name
        self._form_parts = form_parts
        FunctionGenerator.__init__(self, rgl_gf_generator)

    def get_function_name(self):
        return self.generate_new_function_name("ask_%s" % self._predicate_name)

    def get_function_category(self):
        return 'UsrWHQ'

    def get_arguments(self):
        return [
            self.get_individual_argument(form_part) for form_part in self._form_parts if self.is_individual(form_part)
        ]

    def get_semantic_content(self):
        result = 'ask_whq %s' % self._predicate_name
        for sort in self._get_individual_sorts(self._form_parts):
            result += ' %s' % sort
        return result

    def _get_individual_sorts(self, form_parts):
        return [self.get_individual_sort(form_part) for form_part in form_parts if self.is_individual(form_part)]

    def get_natural_language_content(self):
        generated_parts = [self._generate_form_part(form_part) for form_part in self._form_parts]
        return 'mkUsr %s' % self.generate_gf_concatenation(generated_parts)

    def _generate_form_part(self, form_part):
        if self.is_individual(form_part):
            return '(mkUtt %s)' % self.get_individual_sort(form_part)
        elif self.is_string(form_part):
            return self.generate_string_as_utterance(form_part)
        else:
            raise GrammarProcessingException("expected an individual or a string but got %s" % form_part)


class RequestFunctionGenerator(FunctionGenerator):
    def __init__(self, files_generator, action, form_parts):
        self._action = action
        self._form_parts = form_parts
        FunctionGenerator.__init__(self, files_generator)

    def get_function_name(self):
        return self.generate_new_function_name("%s_request" % self._action)

    def get_function_category(self):
        return 'UsrRequest'

    def get_arguments(self):
        return [
            self.get_individual_argument(form_part) for form_part in self._form_parts if self.is_individual(form_part)
        ]

    def get_semantic_content(self):
        result = 'request (pp "%s")' % self._action
        for sort in self._get_individual_sorts(self._form_parts):
            result += ' %s' % sort
        return result

    def _get_individual_sorts(self, form_parts):
        return [self.get_individual_sort(form_part) for form_part in form_parts if self.is_individual(form_part)]

    def get_natural_language_content(self):
        generated_parts = [self._generate_form_part(form_part) for form_part in self._form_parts]
        return 'mkUsr %s' % self.generate_gf_concatenation(generated_parts)

    def _generate_form_part(self, form_part):
        if self.is_individual(form_part):
            return '(mkUtt %s)' % self.get_individual_sort(form_part)
        elif self.is_string(form_part):
            return self.generate_string_as_utterance(form_part)
        else:
            raise GrammarProcessingException("expected an individual or a string but got %s" % form_part)


class ReportFunctionGenerator(FunctionGenerator):
    def __init__(self, files_generator, action, form):
        FunctionGenerator.__init__(self, files_generator)
        self._action = action
        self._form_parts = self.get_utterance_content(form)

    def get_arguments(self):
        return [
            Argument(name=self.get_individual_predicate(form_part), category='SysAnswer')
            for form_part in self._form_parts if self.is_individual_with_predicate(form_part)
        ]

    def get_parameter_list(self):
        predicates = [
            self.get_individual_predicate(form_part) for form_part in self._form_parts
            if self.is_individual_with_predicate(form_part)
        ]
        return self.generate_list(predicates)

    def get_natural_language_content(self):
        generated_parts = [self._generate_form_part(form_part) for form_part in self._form_parts]
        return 'mkSys %s' % self.generate_gf_concatenation(generated_parts)

    def _generate_form_part(self, form_part):
        if self.is_individual_with_predicate(form_part):
            return self.get_individual_predicate(form_part)
        elif self.is_string(form_part):
            return self.generate_string_as_utterance(form_part)
        else:
            raise GrammarProcessingException(
                "expected an individual with a predicate or a string but got %s" % form_part
            )


class ReportEndedFunctionGenerator(ReportFunctionGenerator):
    def get_function_name(self):
        return 'report_ended_%s' % self._action

    def get_function_category(self):
        return 'SysReportEnded'

    def get_semantic_content(self):
        return 'report_ended "%s" %s' % (self._action, self.get_parameter_list())


class ReportStartedFunctionGenerator(ReportFunctionGenerator):
    def get_function_name(self):
        return 'report_started_%s' % self._action

    def get_function_category(self):
        return 'SysReportStarted'

    def get_semantic_content(self):
        return 'report_started "%s" %s' % (self._action, self.get_parameter_list())


class FunctionGeneratorForReportFailedWithUnknownFailure(FunctionGenerator):
    def __init__(self, files_generator, action, form):
        FunctionGenerator.__init__(self, files_generator)
        self._action = action
        self._form_parts = self.get_utterance_content(form)

    def get_function_name(self):
        return 'report_failed_%s_undefined_failure' % self._action

    def get_function_category(self):
        return 'SysReportFailed'

    def get_arguments(self):
        return [
            Argument(name=self.get_individual_predicate(form_part), category='SysAnswer')
            for form_part in self._form_parts if self.is_individual_with_predicate(form_part)
        ]

    def get_semantic_content(self):
        return 'report_failed "%s" %s "%s"' % (
            self._action, self._get_parameter_list(), UNDEFINED_SERVICE_ACTION_FAILURE
        )

    def _get_parameter_list(self):
        predicates = [
            self.get_individual_predicate(form_part) for form_part in self._form_parts
            if self.is_individual_with_predicate(form_part)
        ]
        return self.generate_list(predicates)

    def get_natural_language_content(self):
        return 'mkSys undefined_service_action_failure'


class FunctionGeneratorForSystemAnswerOfCustomSort(FunctionGenerator):
    def __init__(self, files_generator, predicate, form):
        FunctionGenerator.__init__(self, files_generator)
        self._predicate = predicate
        self._form = form

    def get_function_name(self):
        return '%s_sys_answer' % self._predicate.get_name()

    def _get_answer_argument(self):
        sort = self._predicate.getSort()
        argument_category = self.get_sort_category(sort.get_name())
        return Argument('individual', argument_category)

    def _is_individual_of_same_predicate_as_answer(self, form_part):
        return self.is_individual_with_predicate(form_part) and self.get_individual_predicate(
            form_part
        ) == self._predicate.get_name()


class FunctionGeneratorForSystemAnswerOfCustomSortWithoutBackground(FunctionGeneratorForSystemAnswerOfCustomSort):
    def __init__(self, files_generator, predicate, form):
        FunctionGeneratorForSystemAnswerOfCustomSort.__init__(self, files_generator, predicate, form)

    def get_function_category(self):
        return 'SysAnswer'

    def get_arguments(self):
        return [self._get_answer_argument()]

    def get_semantic_content(self):
        return 'pp "%s" individual' % self._predicate.get_name()

    def get_natural_language_content(self):
        if self._form is None:
            return self._default_natural_language_content()
        else:
            form_parts = self.get_utterance_content(self._form)
            generated_parts = [self._generate_form_part(form_part) for form_part in form_parts]
            return 'mkSysAnswer %s' % self.generate_gf_concatenation(generated_parts)

    def _default_natural_language_content(self):
        return 'mkSysAnswer individual'

    def _generate_form_part(self, form_part):
        if self._is_individual_of_same_predicate_as_answer(form_part):
            return '(mkUtt individual)'
        elif self.is_string(form_part):
            return self.generate_string_as_utterance(form_part)
        else:
            raise GrammarProcessingException(
                "expected individual of predicate %s or a string but got %r" % (self._predicate, form_part)
            )


class FunctionGeneratorForSystemAnswerOfCustomSortWithBackground(FunctionGeneratorForSystemAnswerOfCustomSort):
    def __init__(self, files_generator, predicate, form):
        FunctionGeneratorForSystemAnswerOfCustomSort.__init__(self, files_generator, predicate, form)
        self._background_predicates = self._get_background_predicates()

    def _get_background_predicates(self):
        form_parts = self.get_utterance_content(self._form)
        return [
            self.get_individual_predicate(form_part) for form_part in form_parts
            if self._is_background_individual(form_part)
        ]

    def _is_background_individual(self, form_part):
        return self.is_background_individual(form_part, self._predicate.get_name())

    def get_function_category(self):
        return 'System'

    def get_arguments(self):
        return [self._get_answer_argument()] + self._get_background_arguments()

    def _get_background_arguments(self):
        return [Argument(name=predicate, category='SysAnswer') for predicate in self._background_predicates]

    def get_semantic_content(self):
        return 'pp "Move" (move "answer" (pp "%s" individual) %s)' % (
            self._predicate.get_name(), self.generate_list(self._background_predicates)
        )

    def get_natural_language_content(self):
        form_parts = self.get_utterance_content(self._form)
        generated_parts = [self._generate_form_part(form_part) for form_part in form_parts]
        return 'mkSys %s' % self.generate_gf_concatenation(generated_parts)

    def _generate_form_part(self, form_part):
        if self._is_individual_of_same_predicate_as_answer(form_part):
            return '(mkUtt individual)'
        elif self._is_background_individual(form_part):
            return self.get_individual_predicate(form_part)
        elif self.is_string(form_part):
            return self.generate_string_as_utterance(form_part)
        else:
            raise GrammarProcessingException("expected individual or a string but got %r" % form_part)


class ValidityGenerator(GrammarProcessor, GfGenerator):
    def __init__(self, files_generator, validator_interface, forms):
        self._files_generator = files_generator
        self._validator_interface = validator_interface
        self._forms = forms

    def generate(self):
        for parameter_subset in self._get_parameter_subsets_for_validity():
            self._potentially_generate_function_for_parameter_subset(parameter_subset)

    def _get_parameter_subsets_for_validity(self):
        parameters = self._validator_interface.parameters
        for subset_size in range(1, len(parameters) + 1):
            for subset in itertools.combinations(parameters, subset_size):
                yield subset

    def _potentially_generate_function_for_parameter_subset(self, parameter_subset):
        form = self._get_form_for_parameter_subset(parameter_subset)
        if form:
            self._generate_function_for_parameter_subset(form, parameter_subset)

    def _generate_function_for_parameter_subset(self, form, parameter_subset):
        validity_name = self._validator_interface.name
        function_name = self._files_generator.generate_new_function_name(validity_name)
        parameter_names = [parameter.name for parameter in parameter_subset]
        semantic_content = 'rejectICM (set (list %s)) "%s"' % (' '.join(parameter_names), validity_name)
        natural_language_content = self._generate_natural_language_content(form)

        self._files_generator.write_function(
            function_name=function_name,
            category="SysICM",
            arguments=[Argument(parameter.name, "SysAnswer") for parameter in parameter_subset],
            semantic_content=semantic_content,
            natural_language_content=natural_language_content
        )

    def _generate_natural_language_content(self, form):
        form_parts = self.get_utterance_content(form)
        generated_parts = [self._generate_form_part(form_part) for form_part in form_parts]
        return 'mkSys %s' % self.generate_gf_concatenation(generated_parts)

    def _generate_form_part(self, form_part):
        if self.is_individual_with_predicate(form_part):
            return self.get_individual_predicate(form_part)
        elif self.is_string(form_part):
            return self.generate_string_as_utterance(form_part)
        else:
            raise GrammarProcessingException(
                "expected an individual with a predicate or a string but got %s" % form_part
            )

    def _get_form_for_parameter_subset(self, parameter_subset):
        for form in self._forms:
            if not self._form_contains_parameter_not_in_subset(form, parameter_subset):
                return form

    def _form_contains_parameter_not_in_subset(self, form, parameter_subset):
        parameter_names_in_form = set(self._parameters_in_form(form))
        parameter_names_in_subset = set([parameter.name for parameter in parameter_subset])
        parameters_in_form_but_not_in_subset = parameter_names_in_form.difference(parameter_names_in_subset)
        return len(parameters_in_form_but_not_in_subset) > 0

    def _parameters_in_form(self, form):
        parameter_names_in_device = [parameter.name for parameter in self._validator_interface.parameters]
        result = []
        for part in form.children:
            if self.is_individual_with_predicate(part):
                parameter_name = self.get_individual_predicate(part)
                if parameter_name in parameter_names_in_device:
                    result.append(parameter_name)
                else:
                    raise GrammarProcessingException(
                        "unexpected parameter name %s in form %s (device declares %s)" %
                        (parameter_name, form, parameter_names_in_device)
                    )
        return result


class Argument:
    def __init__(self, name, category):
        self.name = name
        self.category = category



class Constants:
    ACTION = "action"
    SYS_ANSWER = "sys-answer"
    POSITIVE_SYS_ANSWER = "positive-sys-answer"
    NEGATIVE_SYS_ANSWER = "negative-sys-answer"
    PREDICATE = "predicate"
    USER_QUESTION = "user-question"
    SYS_QUESTION = "sys-question"
    ISSUE_TITLE = "issue-title"
    ACTION_TITLE = "action-title"
    ANSWER_COMBINATION = "answer-combination"
    REPORT_ENDED = "report-ended"
    REPORT_STARTED = "report-started"
    REPORT_FAILED = "report-failed"
    PREREPORT = "prereport"
    PRECONFIRM = "preconfirm"
    VALIDITY = "validity"
    INDIVIDUAL = "individual"
    STRING = "string"
    GREETING = "greeting"
    GRAMMAR = "grammar"
    SLOT = "slot"
    ONE_OF = "one-of"
    ITEM = "item"
    NP = "np"
    INDEFINITE = "indefinite"
    DEFINITE = "definite"
    VP = "vp"
    INFINITIVE = "infinitive"
    IMPERATIVE = "imperative"
    ING_FORM = "ing-form"
    OBJECT = "object"


class Node:
    def __init__(self, type_, parameters=None, children=None):
        self.type = type_
        if parameters is None:
            self.parameters = {}
        else:
            self.parameters = parameters
        if children is None:
            self.children = []
        else:
            self.children = children

    def __str__(self):
        return self._pretty_string()

    def _pretty_string(self, indentation=0):
        string = "%sNode(%s, %s)" % ("  " * indentation, self.type, self.parameters)
        if isinstance(self.children, list):
            if len(self.children) > 0:
                for child in self.children:
                    if isinstance(child, Node):
                        string += "\n%s" % (child._pretty_string(indentation + 1))
                    else:
                        string += "\n%s%r" % ("  " * (indentation + 1), child)
            else:
                string += " %s" % self.children
        else:
            string += "\n%s%s" % ("  " * (indentation + 1), self.children)
        return string

    def __repr__(self):
        return "Node(%s, %s, %r)" % (self.type, self.parameters, self.children)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return other.type == self.type and other.parameters == self.parameters and other.children == self.children
        else:
            return False

    def add_child(self, child):
        self.children.append(child)

    def get_child(self, type_, parameters={}):
        for child in self.children:
            if child.type == type_ and child.parameters == parameters:
                return child

    def get_single_child(self):
        if len(self.children) == 1:
            return self.children[0]
        else:
            raise Exception("expected this node to have a single child:\n%s" % self)



class AbstractBuildStep(object):
    def __init__(self, ddd, language_codes, ddd_root_directory, grammar_directory, verbose):
        self._ddd = ddd
        self._name = ddd.name
        self._language_codes = language_codes
        self._ddd_root_directory = ddd_root_directory
        self._grammar_directory = grammar_directory
        self._verbose = verbose

    def build(self):
        raise NotImplementedError()

    def _directory_of_rasa_files(self, language_code):
        return "%s/%s" % ("build_rasa", language_code)

    def _log_command(self, cmd):
        if self._verbose:
            print("calling %r" % cmd)



from tala.model.semantic_object import OntologySpecificSemanticObject


class Action(OntologySpecificSemanticObject):
    def __init__(self, value, ontology_name):
        OntologySpecificSemanticObject.__init__(self, ontology_name)
        self.value = value

    def is_action(self):
        return True

    def isRespondPlanItem(self):
        return False

    def get_value(self):
        return self.value

    def is_top_action(self):
        return self.value == "top"

    def is_up_action(self):
        return self.value == "up"

    def __unicode__(self):
        return self.value

    def __hash__(self):
        return hash((self.ontology_name, self.value))

    def __eq__(self, other):
        try:
            return other.is_action() and other.get_value() == self.get_value(
            ) and other.ontology_name == self.ontology_name
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)


class TopAction(Action):
    def __init__(self, ontology_name):
        Action.__init__(self, "top", ontology_name)


class UpAction(Action):
    def __init__(self, ontology_name):
        Action.__init__(self, "up", ontology_name)



class UnicodifyException(Exception):
    pass


def unicodify(object):
    if isinstance(object, (str, unicode)):
        return u"'%s'" % unicode(object)
    if isinstance(object, list):
        return _sequence_string(object)
    if isinstance(object, dict):
        return _dict_string(object)
    if isinstance(object, tuple):
        return _tuple_string(object)
    else:
        try:
            return unicode(object)
        except UnicodeDecodeError as exception:
            raise UnicodifyException("failed to unicodify object of class %s (%s)" % (object.__class__, exception))


def _sequence_string(sequence):
    return u"[%s]" % u", ".join([unicodify(item) for item in sequence])


def _dict_string(dict_):
    formatted_key_value_pairs = [(unicodify(key), unicodify(value)) for key, value in dict_.iteritems()]
    formatted_content = [u"%s: %s" % pair for pair in formatted_key_value_pairs]
    return u"{%s}" % u", ".join(formatted_content)


def _tuple_string(_tuple):
    return u"(%s)" % u", ".join([unicodify(item) for item in _tuple])



HEADER = '\033[95m'
OKBLUE = '\033[94m'
OKGREEN = '\033[92m'
WARNING = '\033[93m'
FAIL = '\033[91m'
ENDC = '\033[0m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'


def header(string):
    return "%s%s%s" % (HEADER, string, ENDC)


def ok_blue(string):
    return "%s%s%s" % (OKBLUE, string, ENDC)


def ok_green(string):
    return "%s%s%s" % (OKGREEN, string, ENDC)


def warning(string):
    return "%s%s%s" % (WARNING, string, ENDC)


def fail(string):
    return "%s%s%s" % (FAIL, string, ENDC)


def bold(string):
    return "%s%s%s" % (BOLD, string, ENDC)


def underline(string):
    return "%s%s%s" % (UNDERLINE, string, ENDC)



import os
import re

from tala.model.date_time import DateTime


class GrammarSanityException(Exception):
    pass


BIND = "&+"
META = "#?#"

TOKENISE_RE = re.compile(r"\s*(p\.m\.|a\.m\.|\d{1,2}\:\d{2}|[\w0-9_\\\-']+|\S)\s*", flags=re.UNICODE)

PREPROCESS_RE_SINGLE_QUOTES = re.compile(r'\'([^\']*)\'')
PREPROCESS_RE_DOUBLE_QUOTES = re.compile(r'"([^"]*)"')
PUNCTUATION_RE = re.compile(r"\s+([!,.?])")

NL_PLACEHOLDER_PREFIX = "nl_placeholder"
SEM_PLACEHOLDER_PREFIX = "sem_placeholder"
PLACEHOLDER_NAME_PREFIX = "placeholder"

UNFORMATTED_PLACEHOLDER_REGEXP = "_%s_\w+(\s|$)"
NL_PLACEHOLDER_REGEXP = UNFORMATTED_PLACEHOLDER_REGEXP % NL_PLACEHOLDER_PREFIX
SEM_PLACEHOLDER_REGEXP = UNFORMATTED_PLACEHOLDER_REGEXP % SEM_PLACEHOLDER_PREFIX

MAX_NUM_PLACEHOLDERS = 10


def tokenise(s):
    tokenised = filter(None, TOKENISE_RE.split(s))
    return tokenised


def preprocess_strings(unprocessed_string, table):
    def replace_and_store(matchobj):
        count = len(table)
        key = '_STR%s_' % count
        table[key] = matchobj.group(1)
        return key

    preprocessed_string = PREPROCESS_RE_SINGLE_QUOTES.sub(replace_and_store, unprocessed_string)
    preprocessed_string = PREPROCESS_RE_DOUBLE_QUOTES.sub(replace_and_store, preprocessed_string)
    return preprocessed_string


def preprocess_integers(unprocessed_string, table):
    def replace_and_store(matchobj):
        count = len(table)
        key = '_integer_placeholder_%s_' % count
        table[key] = matchobj.group(1)
        return "(%s)" % key

    return re.sub("\((\d{3,})\)", replace_and_store, unprocessed_string)


def preprocess_datetimes(unprocessed_string, table):
    def replace_and_store(matchobj):
        count = len(table)
        key = '_datetime_placeholder_%s_' % count
        iso8601_string = matchobj.group(1)
        date_time = DateTime(iso8601_string)
        table[key] = date_time.human_standard()
        return "(%s)" % key

    return re.sub("\(datetime\(([^()]+)\)\)", replace_and_store, unprocessed_string)


def replace_placeholders(string, table):
    result = string
    for placeholder, replacement in table.iteritems():
        result = result.replace(placeholder, replacement)
    return result


def str_lin(lin, marked=None):
    return " ; ".join(str_phrase(phrase, marked) for phrase in lin)


def lin_to_moves(lin):
    moves = [move.strip() for move in lin_to_move(lin).split(";") if META not in move]
    return moves


def lin_to_move(lin):
    return "".join(concat_string_tokens(lin))


def concat_string_tokens(tokens):
    try:
        first_quote_index = tokens.index('"')
        tokens_from_first_quote = tokens[first_quote_index + 1:]
        second_quote_index = tokens_from_first_quote.index('"')
    except ValueError:
        return tokens
    tokens_between_quotes = tokens_from_first_quote[0:second_quote_index]
    string_token = " ".join(tokens_between_quotes)
    tokens_before_first_quote = tokens[0:first_quote_index]
    tokens_after_second_quote = tokens_from_first_quote[second_quote_index + 1:]
    result = tokens_before_first_quote
    result.append('"')
    result.append(string_token)
    result.append('"')
    result.extend(concat_string_tokens(tokens_after_second_quote))
    return result


def str_phrase(phrase, marked=None):
    phrase = bind_tokens(phrase)
    str = " ".join(str_token(token, marked) for token in phrase)
    str = PUNCTUATION_RE.sub(r"\1", str)
    return str


def str_token(token, marked=None):
    if isinstance(token, basestring):
        return token
    word, path = token
    star = "*" if is_path(marked) and startswith(path, marked) else ""
    return star + word + "/" + str_path(path)


def is_path(path):
    return (isinstance(path, tuple) and all(isinstance(n, int) and n >= 1 for n in path))


def bind_tokens(tokens):
    result = []
    binding = False
    for token in tokens:
        if binding:
            result[-1] += token
            binding = False
        elif token == BIND:
            binding = True
        else:
            result.append(token)
    return result


def startswith(sequence, prefix):
    return sequence[:len(prefix)] == prefix


def str_path(path):
    if path is None:
        return unicode(None)
    return "".join(map(str, path))


def assert_grammar_is_lower_case(filename):
    if not os.path.exists(filename):
        return
    with open(filename) as f:
        for line_number, line in enumerate(f, start=1):
            line = line.strip()
            line = without_comments(line)
            expected_lower_case_tokens = language_tokens(line)
            for expected_lower_case_text in expected_lower_case_tokens:
                if not expected_lower_case_text == expected_lower_case_text.lower():
                    msg_string = "Expected lower case grammar %r, but found deviation on line %s: %r"
                    msg = msg_string % (filename, line_number, expected_lower_case_text)
                    raise GrammarSanityException(msg)


def language_token_is_constant(entry):
    return entry.startswith("_") and entry.endswith("_")


def language_token_is_nl_placeholder(token):
    return re.search(NL_PLACEHOLDER_REGEXP, token) is not None


def remove_nl_placeholders_from_string(string):
    return re.sub(NL_PLACEHOLDER_REGEXP, "", string)


def remove_sem_placeholders_from_string(string):
    return re.sub(SEM_PLACEHOLDER_REGEXP, "", string)


def nl_user_answer_placeholder_of_sort(sort, index):
    return "_%s_%s%s_" % (NL_PLACEHOLDER_PREFIX, sort, index)


def semantic_user_answer_placeholder_of_sort(sort, index):
    return "_%s_%s%s_" % (SEM_PLACEHOLDER_PREFIX, sort, index)


def name_of_user_answer_placeholder_of_sort(sort, index):
    return "%s_%s%s" % (PLACEHOLDER_NAME_PREFIX, sort, index)


def without_comments(line):
    return line.split("--")[0]


def language_tokens(line):
    tokens = line.split('"')
    language_tokens = tokens[1::2]  # every odd element
    return [token for token in language_tokens if not language_token_is_constant(token)]


def lower_gf_string_but_insert_capit_and_bind(string):
    lowered_string = '"'
    previous_char = None
    for char in string:
        if not char.isupper():
            lowered_string += char
        else:
            is_char_in_mid_of_word = previous_char and not previous_char.isspace()
            if is_char_in_mid_of_word:
                lowered_string += '" ++ CAPIT ++ BIND ++ "%s' % char.lower()
            else:
                lowered_string += '" ++ CAPIT ++ "%s' % char.lower()

        previous_char = char
    lowered_string += '"'
    if "CAPIT" in lowered_string:
        return "(%s)" % lowered_string
    return lowered_string



from pathlib import Path
import pytest

from tala.cli import console_script
from tala.utils.chdir import chdir


class TestDDDs(object):
    @pytest.mark.parametrize(
        "backend_config", [
            "hello_world.json",
            "http_service.config.json",
            "incremental_search.config.json",
            "literals_grammar.json",
            "mockup_travel.json",
            "partial_parsing.json",
            "rasa_answers.config.json",
            "rasa_for_dynamic_entities.config.json",
            "rasa_for_rgl.config.json",
            "rasa_for_static_entities.config.json",
            "rasa_numbers.config.json",
            "rasa_strings.config.json",
            "rasa_time.config.json",
            "send_to_frontend.config.json",
            "small_grammar.json",
            "tidePooler.json",
        ]
    )
    def test_verify_on_ddd(self, backend_config):
        with self._given_changed_to_ddd_directory():
            self._when_verifying_with(backend_config)
            self._then_no_errors_occured()

    def _given_changed_to_ddd_directory(self):
        directory = Path("tala") / "ddds"
        return chdir(directory)

    def _when_verifying_with(self, backend_config):
        console_script.main([
            "verify", "--config", backend_config, "--rasa-config", "duckling_enabled.rasa.config.json"
        ])

    def _then_no_errors_occured(self):
        pass

    def test_verify_on_multiple_ddds(self):
        with self._given_changed_to_multi_ddd_directory():
            self._when_verifying()
            self._then_no_errors_occured()

    def _given_changed_to_multi_ddd_directory(self):
        directory = Path("tala") / "ddds" / "multi_ddd"
        return chdir(directory)

    def _when_verifying(self):
        console_script.main(["verify"])



NOUN_PHRASE = "noun-phrase"
NOUN = "noun"
VERB_PHRASE = "verb-phrase"
VERB = "verb"
LEXICON = "lexicon"
PROPER_NOUN = "proper-noun"
UTTERANCE = "utterance"
REQUEST = "request"
PREDICATE = "predicate"
SINGULAR = "singular"



import copy

from tala.model.lambda_abstraction import LambdaAbstractedGoalProposition, LambdaAbstractedPredicateProposition
from tala.model.test.utils import LibTestCase


class LambdaAbstractedGoalPropositionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.lambda_prop = LambdaAbstractedGoalProposition()

    def test_unicode(self):
        self.assertEquals("X.goal(X)", unicode(self.lambda_prop))

    def test_equality(self):
        lambda_prop1 = LambdaAbstractedGoalProposition()
        lambda_prop2 = LambdaAbstractedGoalProposition()
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(lambda_prop1, lambda_prop2)

    def test_inequality(self):
        lambda_prop = LambdaAbstractedGoalProposition()
        non_identical_lambda_prop = self.lambda_abstracted_price_prop
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(lambda_prop, non_identical_lambda_prop)

    def test_hashing(self):
        set([self.lambda_prop])

    def test_is_lambda_abstracted_goal_proposition(self):
        self.assertTrue(self.lambda_abstracted_goal_prop.is_lambda_abstracted_goal_proposition())
        self.assertFalse(self.lambda_abstracted_dest_city_prop.is_lambda_abstracted_goal_proposition())


class LambdaAbstractedPredicatePropositionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.lambda_prop = LambdaAbstractedPredicateProposition(self.predicate_dest_city, self.ontology.get_name())

    def test_unicode(self):
        self.assertEquals("X.dest_city(X)", unicode(self.lambda_prop))

    def test_equality(self):
        lambda_prop = self.lambda_abstracted_dest_city_prop
        identical_lambda_prop = copy.copy(self.lambda_abstracted_dest_city_prop)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(lambda_prop, identical_lambda_prop)

    def test_inequality(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(
            self.lambda_abstracted_dest_city_prop, self.lambda_abstracted_price_prop
        )

    def test_not_equals_proposition(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(
            self.proposition_dest_city_paris, self.lambda_prop
        )

    def test_hashing(self):
        set([self.lambda_prop])

    def test_is_lambda_abstracted_predicate_proposition(self):
        self.assertTrue(self.lambda_abstracted_dest_city_prop.is_lambda_abstracted_predicate_proposition())
        self.assertFalse(self.lambda_abstracted_goal_prop.is_lambda_abstracted_predicate_proposition())



import copy

from tala.model.error import OntologyError
from tala.model.polarity import Polarity
from tala.model.semantic_object import SemanticObject, OntologySpecificSemanticObject, SemanticObjectWithContent
from tala.utils.unicodify import unicodify


class PropositionsFromDifferentOntologiesException(Exception):
    pass


class Proposition(SemanticObject):
    UNDERSTANDING = "UNDERSTANDING"
    GOAL = "GOAL"
    PREDICATE = "PREDICATE"
    PRECONFIRMATION = "PRECONFIRMATION"
    PREREPORT = "PREREPORT"
    MUTE = "MUTE"
    UNMUTE = "UNMUTE"
    QUIT = "QUIT"
    PREDICTED = "PREDICTED"
    REJECTED = "REJECTED"
    RESOLVEDNESS = "RESOLVEDNESS"
    SERVICE_RESULT = "SERVICE_RESULT"
    SERVICE_ACTION_STARTED = "SERVICE_ACTION_STARTED"
    SERVICE_ACTION_TERMINATED = "SERVICE_ACTION_TERMINATED"
    PROPOSITION_SET = "PROPOSITION_SET"

    def __init__(self, type, polarity=None):
        SemanticObject.__init__(self)
        self._type = type
        if polarity is None:
            polarity = Polarity.POS
        self._polarity = polarity
        self._predicted = False

    def __eq__(self, other):
        try:
            return other.is_proposition() and self.get_type() == other.get_type() and self.get_polarity(
            ) == other.get_polarity()
        except AttributeError:
            return False

    def get_polarity_prefix_string(self):
        if self._polarity == Polarity.NEG:
            return "~"
        else:
            return ""

    def get_type(self):
        return self._type

    def is_proposition(self):
        return True

    def is_understanding_proposition(self):
        return self._type == Proposition.UNDERSTANDING

    def is_predicate_proposition(self):
        return self._type == Proposition.PREDICATE

    def is_preconfirmation_proposition(self):
        return self._type == Proposition.PRECONFIRMATION

    def is_prereport_proposition(self):
        return self._type == Proposition.PREREPORT

    def is_service_result_proposition(self):
        return self._type == Proposition.SERVICE_RESULT

    def is_mute_proposition(self):
        return self._type == Proposition.MUTE

    def is_unmute_proposition(self):
        return self._type == Proposition.UNMUTE

    def is_quit_proposition(self):
        return self._type == Proposition.QUIT

    def is_goal_proposition(self):
        return self._type == Proposition.GOAL

    def is_rejected_proposition(self):
        return self._type == Proposition.REJECTED

    def is_resolvedness_proposition(self):
        return self._type == Proposition.RESOLVEDNESS

    def is_service_action_started_proposition(self):
        return self._type == Proposition.SERVICE_ACTION_STARTED

    def is_service_action_terminated_proposition(self):
        return self._type == Proposition.SERVICE_ACTION_TERMINATED

    def is_predicted_proposition(self):
        return self._type == Proposition.PREDICTED

    def get_polarity(self):
        return self._polarity

    def is_positive(self):
        return self._polarity == Polarity.POS

    def is_predicted(self):
        return self._predicted

    def negate(self):
        proposition_copy = copy.copy(self)
        if proposition_copy.is_positive():
            proposition_copy._polarity = Polarity.NEG
        else:
            proposition_copy._polarity = Polarity.POS
        return proposition_copy

    def is_incompatible_with(self, other):
        return False

    def __repr__(self):
        return "%s%s" % (Proposition.__name__, (self._type, self._polarity))


class PropositionWithSemanticContent(Proposition, SemanticObjectWithContent):
    def __init__(self, type, content, polarity=None):
        Proposition.__init__(self, type, polarity)
        SemanticObjectWithContent.__init__(self, content)
        self._content = content

    @property
    def content(self):
        return self._content

    def __eq__(self, other):
        try:
            return other.is_proposition() and other.has_semantic_content(
            ) and self.content == other.content and self.get_type() == other.get_type() and self.get_polarity(
            ) == other.get_polarity()
        except AttributeError:
            return False

    def __repr__(self):
        return "%s%s" % (PropositionWithSemanticContent.__name__, (self._type, self.content, self._polarity))


class ControlProposition(Proposition):
    def __init__(self, type, polarity=None):
        Proposition.__init__(self, type, polarity)

    def __ne__(self, other):
        return not self == other

    def __unicode__(self):
        return self._type

    def __hash__(self):
        return self._type.__hash__()


class MuteProposition(ControlProposition):
    def __init__(self=None):
        Proposition.__init__(self, Proposition.MUTE)


class UnmuteProposition(ControlProposition):
    def __init__(self=None):
        Proposition.__init__(self, Proposition.UNMUTE)


class QuitProposition(ControlProposition):
    def __init__(self=None):
        Proposition.__init__(self, Proposition.QUIT)


class PredicateProposition(PropositionWithSemanticContent):
    def __init__(self, predicate, individual=None, polarity=None, predicted=False):
        if polarity is None:
            polarity = Polarity.POS
        PropositionWithSemanticContent.__init__(self, Proposition.PREDICATE, predicate, polarity=polarity)
        self.predicate = predicate
        self.individual = individual
        self._predicted = predicted
        if individual is not None and individual.getSort() != predicate.getSort():
            raise OntologyError(("Sortal mismatch between predicate %s " + "(sort %s) and individual %s (sort %s)") %
                                (predicate, predicate.getSort(), individual, individual.getSort()))

    def getPredicate(self):
        return self.predicate

    def getArgument(self):
        return self.individual

    def is_incompatible_with(self, other):
        return (
            self.negate() == other or (
                self._proposes_other_individual_of_same_predicate(other)
                and not self.getPredicate().allows_multiple_instances()
            ) or self._is_feature_of_the_following_negative_proposition(other)
            or (self._polarity == Polarity.NEG and self._is_feature(other))
        )

    def _proposes_other_individual_of_same_predicate(self, other):
        return (
            other.is_predicate_proposition() and self.getPredicate() == other.getPredicate() and self.is_positive()
            and other.is_positive() and self.getArgument() != other.getArgument()
        )

    def _is_feature_of_the_following_negative_proposition(self, other):
        return (
            other.is_predicate_proposition() and self.predicate.is_feature_of(other.getPredicate())
            and other.get_polarity() == Polarity.NEG
        )

    def _is_feature(self, answer):
        return (answer.is_predicate_proposition() and answer.getPredicate().is_feature_of(self.predicate))

    def __eq__(self, other):
        try:
            return other is not None and other.is_proposition() and other.is_predicate_proposition(
            ) and self.getPredicate() == other.getPredicate() and self.getArgument() == other.getArgument(
            ) and self.get_polarity() == other.get_polarity()
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __unicode__(self):
        if self.individual is None:
            return "%s%s" % (self.get_polarity_prefix_string(), unicode(self.predicate))
        else:
            return "%s%s(%s)" % (self.get_polarity_prefix_string(), unicode(self.predicate), unicode(self.individual))

    def __hash__(self):
        return hash((self.predicate, self.individual, self.get_polarity()))

    def __repr__(self):
        return "%s%s" % (self.__class__.__name__, (self.predicate, self.individual, self._polarity, self._predicted))


class GoalProposition(PropositionWithSemanticContent):
    def __init__(self, goal, polarity=None):
        PropositionWithSemanticContent.__init__(self, Proposition.GOAL, goal, polarity=polarity)
        self._goal = goal

    def get_goal(self):
        return self._goal

    def __eq__(self, other):
        try:
            return other.is_proposition() and other.is_goal_proposition() and self._goal == other.get_goal(
            ) and self._polarity == other.get_polarity()
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __unicode__(self):
        return "%sgoal(%s)" % (self.get_polarity_prefix_string(), self._goal)

    def __hash__(self):
        return hash((self._goal, self._polarity))


class PreconfirmationProposition(Proposition, OntologySpecificSemanticObject):
    def __init__(self, ontology_name, service_action, arguments, polarity=None):
        self.service_action = service_action
        self._arguments = arguments
        self._hash = hash((ontology_name, polarity, service_action, frozenset(arguments)))
        Proposition.__init__(self, Proposition.PRECONFIRMATION, polarity)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    def get_service_action(self):
        return self.service_action

    def get_arguments(self):
        argument_list = []
        for param in self._arguments:
            argument_list.append(param)
        return argument_list

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_preconfirmation_proposition()
                and other.ontology_name == self.ontology_name and other.get_type() == self.get_type()
                and other.get_service_action() == self.get_service_action()
                and other.get_polarity() == self.get_polarity() and other.get_arguments() == self.get_arguments()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __unicode__(self):
        return "%spreconfirmed(%s, %s)" % (
            self.get_polarity_prefix_string(), unicode(self.service_action), unicodify(self._arguments)
        )

    def __hash__(self):
        return self._hash


class PrereportProposition(Proposition, OntologySpecificSemanticObject):
    def __init__(self, ontology_name, service_action, argument_list):
        self.service_action = service_action
        self.argument_set = frozenset(argument_list)
        Proposition.__init__(self, Proposition.PREREPORT)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    def get_service_action(self):
        return self.service_action

    def get_arguments(self):
        argument_list = []
        for param in self.argument_set:
            argument_list.append(param)
        return argument_list

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_prereport_proposition()
                and other.ontology_name == self.ontology_name
                and other.get_service_action() == self.get_service_action()
                and other.get_polarity() == self.get_polarity() and other.argument_set == self.argument_set
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.ontology_name, self.service_action, self.argument_set))

    def __unicode__(self):
        return "prereported(%s, %s)" % (unicode(self.service_action), unicode(self.get_arguments()))


class ServiceActionTerminatedProposition(Proposition, OntologySpecificSemanticObject):
    def __init__(self, ontology_name, service_action, polarity=None):
        self.service_action = service_action
        Proposition.__init__(self, Proposition.SERVICE_ACTION_TERMINATED, polarity)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    def get_service_action(self):
        return self.service_action

    def __unicode__(self):
        return "%sservice_action_terminated(%s)" % (self.get_polarity_prefix_string(), unicode(self.service_action))

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_service_action_terminated_proposition()
                and other.ontology_name == self.ontology_name
                and other.get_service_action() == self.get_service_action()
                and other.get_polarity() == self.get_polarity()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.service_action)

    def __repr__(self):
        return "%s(%r, %r, %r)" % (
            self.__class__.__name__, self.ontology_name, self.service_action, self.get_polarity()
        )


class PredictedProposition(PropositionWithSemanticContent):
    def __init__(self, propositional_answer, polarity=None):
        self.predicted_proposition = propositional_answer
        PropositionWithSemanticContent.__init__(self, Proposition.PREDICTED, propositional_answer, polarity)

    def get_prediction(self):
        return self.predicted_proposition

    def __unicode__(self):
        return "%spredicted(%s)" % (self.get_polarity_prefix_string(), self.predicted_proposition)

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_predicted_proposition()
                and other.predicted_proposition == self.predicted_proposition
                and other.get_polarity() == self.get_polarity()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.predicted_proposition)


class ServiceResultProposition(Proposition, OntologySpecificSemanticObject):
    STARTED = "STARTED"
    ENDED = "ENDED"

    def __init__(self, ontology_name, service_action, arguments, result, status=None):
        self.service_action = service_action
        self.arguments = arguments
        self.result = result
        self._manage_status(status)
        Proposition.__init__(self, Proposition.SERVICE_RESULT)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    def _manage_status(self, status):
        if not status:
            self.status = ServiceResultProposition.ENDED
        else:
            self.status = status

    def get_service_action(self):
        return self.service_action

    def get_arguments(self):
        return self.arguments

    def get_result(self):
        return self.result

    def get_status(self):
        return self.status

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_service_result_proposition()
                and other.get_service_action() == self.get_service_action()
                and other.get_arguments() == self.get_arguments() and other.get_result() == self.get_result()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __unicode__(self):
        return "ServiceResultProposition(%s, %s, %s)" % (self.service_action, unicodify(self.arguments), self.result)

    def __hash__(self):
        return hash((self.get_service_action(), self.get_result()))


class ServiceActionStartedProposition(Proposition, OntologySpecificSemanticObject):
    def __init__(self, ontology_name, service_action, parameters=[]):
        self.service_action = service_action
        self.parameters = parameters
        Proposition.__init__(self, Proposition.SERVICE_ACTION_STARTED)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_service_action_started_proposition()
                and other.ontology_name == self.ontology_name and other.service_action == self.service_action
                and other.get_polarity() == self.get_polarity()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __unicode__(self):
        return "ServiceActionStartedProposition(%s, %s)" % (self.service_action, unicodify(self.parameters))


class RejectedPropositions(PropositionWithSemanticContent):
    def __init__(self, rejected_combination, polarity=None, reason=None):
        self.rejected_combination = rejected_combination
        self.reason_for_rejection = reason
        PropositionWithSemanticContent.__init__(self, Proposition.REJECTED, rejected_combination, polarity)

    def get_rejected_combination(self):
        return self.rejected_combination

    def get_reason(self):
        return self.reason_for_rejection

    def __unicode__(self):
        if self.reason_for_rejection:
            return "%srejected(%s, %s)" % (
                self.get_polarity_prefix_string(), self.rejected_combination, self.reason_for_rejection
            )
        else:
            return "%srejected(%s)" % (self.get_polarity_prefix_string(), unicode(self.rejected_combination))

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_rejected_proposition()
                and other.get_rejected_combination() == self.get_rejected_combination()
                and other.get_reason() == self.get_reason() and other.get_polarity() == self.get_polarity()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)


class PropositionSet(Proposition):
    def __init__(self, propositions, polarity=Polarity.POS):
        self._propositions = propositions
        self._hash = hash(frozenset(propositions))
        Proposition.__init__(self, Proposition.PROPOSITION_SET, polarity)

    def is_proposition_set(self):
        return True

    def __iter__(self):
        return self._propositions.__iter__()

    def __hash__(self):
        return self._hash

    def get_propositions(self):
        return self._propositions

    def unicode_propositions(self):
        return u"[%s]" % u", ".join([unicode(proposition) for proposition in self._propositions])

    def is_single_alt(self):
        return len(self._propositions) == 1

    def is_multi_alt(self):
        return len(self._propositions) > 1

    def get_single_alt(self):
        for proposition in self._propositions:
            return proposition

    def is_goal_alts(self):
        for proposition in self._propositions:
            if proposition.is_goal_proposition():
                return True
        return False

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_proposition_set() and self.get_polarity() == other.get_polarity()
                and self.get_propositions() == other.get_propositions()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __unicode__(self):
        return "%sset(%s)" % (self.get_polarity_prefix_string(), self.unicode_propositions())

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self._propositions, self._polarity)

    def getPredicate(self):
        predicates = set([proposition.getPredicate() for proposition in self._propositions])
        if len(predicates) == 1:
            return list(predicates)[0]
        else:
            raise Exception("cannot get predicate for proposition set with zero or mixed predicates")

    def is_ontology_specific(self):
        if not all([proposition.is_ontology_specific() for proposition in self._propositions]):
            return False
        if not any([proposition.is_ontology_specific() for proposition in self._propositions]):
            return False
        return self._all_propositions_from_same_ontology()

    def _all_propositions_from_same_ontology(self):
        first_ontology_name = self._ontology_names[0]
        return all([name == first_ontology_name for name in self._ontology_names])

    @property
    def ontology_name(self):
        if self.is_ontology_specific() and self._all_propositions_from_same_ontology:
            return self._propositions[0].ontology_name

        message = "Expected all propositions %s\n\nin ontology %s\n\nbut they're from %s" % (
            self._propositions, self._propositions[0].ontology_name, self._ontology_names
        )
        raise PropositionsFromDifferentOntologiesException(message)

    @property
    def _ontology_names(self):
        ontology_names = [proposition.ontology_name for proposition in self._propositions]
        return ontology_names

    def has_semantic_content(self):
        return True


class UnderstandingProposition(PropositionWithSemanticContent):
    def __init__(self, speaker, content, polarity=Polarity.POS):
        PropositionWithSemanticContent.__init__(self, Proposition.UNDERSTANDING, content, polarity=polarity)
        self._speaker = speaker
        self._content = content

    def get_speaker(self):
        return self._speaker

    def get_content(self):
        return self._content

    def __unicode__(self):
        return "und(%s, %s)" % (unicode(self._speaker), unicode(self._content))

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_understanding_proposition()
                and self.get_speaker() == other.get_speaker() and self.get_content() == other.get_content()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self._content, self._speaker))


class ResolvednessProposition(PropositionWithSemanticContent):
    def __init__(self, issue):
        self.issue = issue
        PropositionWithSemanticContent.__init__(self, Proposition.RESOLVEDNESS, issue)

    def get_issue(self):
        return self.issue

    def __unicode__(self):
        return "resolved(%s)" % self.issue

    def __eq__(self, other):
        try:
            return (
                other.is_proposition() and other.is_resolvedness_proposition() and other.get_issue() == self.get_issue()
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.get_issue())



import copy

from tala.model.error import DomainError
from tala.model.plan_item import PlanItemWithSemanticContent, PlanItem


class QuestionRaisingPlanItem(PlanItemWithSemanticContent):
    GRAPHICAL_TYPE_TEXT = "text"
    GRAPHICAL_TYPE_LIST = "list"
    SOURCE_SERVICE = "service"
    SOURCE_DOMAIN = "domain"
    ALPHABETIC = "alphabetic"

    def __init__(self, domain_name, type, content):
        if not content.is_question():
            raise DomainError("cannot create QuestionRaisingPlanItem " + "from non-question %s" % content)
        self._domain_name = domain_name
        PlanItemWithSemanticContent.__init__(self, type, content)

    @property
    def domain_name(self):
        return self._domain_name

    def is_question_raising_item(self):
        return True

    def get_question(self):
        return self.getContent()

    def clone_as_type(self, type):
        clone = copy.deepcopy(self)
        clone._type = type
        return clone

    def __str__(self):
        if self._content is None:
            content_string = ""
        else:
            content_string = unicode(self._content)
        return "%s(%s)" % (unicode(self._type), content_string)


class FindoutPlanItem(QuestionRaisingPlanItem):
    def __init__(self, domain_name, content):
        QuestionRaisingPlanItem.__init__(self, domain_name, PlanItem.TYPE_FINDOUT, content)


class RaisePlanItem(QuestionRaisingPlanItem):
    def __init__(self, domain_name, content):
        QuestionRaisingPlanItem.__init__(self, domain_name, PlanItem.TYPE_RAISE, content)


class UnexpectedDomainException(Exception):
    pass


class QuestionRaisingPlanItemOfDomain(object):
    def __init__(self, domain, plan_item):
        if domain.name != plan_item.domain_name:
            raise UnexpectedDomainException(
                "Expected domain '%s' to match domain of plan item %s but it was '%s'" %
                (plan_item.domain_name, plan_item, domain)
            )
        self._domain = domain
        self._plan_item = plan_item

    def get_alternatives(self):
        return self._domain.get_alternatives(self._plan_item.get_question())

    def get_graphical_type(self):
        return self._domain.get_graphical_type(self._plan_item.get_question())

    def get_incremental(self):
        return self._domain.get_incremental(self._plan_item.get_question())

    def get_source(self):
        return self._domain.get_source(self._plan_item.get_question())

    def get_format(self):
        return self._domain.get_format(self._plan_item.get_question())

    def get_default(self):
        return self._domain.get_default(self._plan_item.get_question())

    def get_service_query(self):
        return self._domain.get_service_query(self._plan_item.get_question())

    def get_label_questions(self):
        return self._domain.get_label_questions(self._plan_item.get_question())

    def has_parameters(self):
        return (
            self.get_alternatives() or self.get_graphical_type() or self.get_incremental() or self.get_source()
            or self.get_format() or self.get_default() or self.get_service_query() or self.get_label_questions()
        )



class ValidationResult(object):
    @property
    def is_successful(self):
        raise NotImplementedError("This property needs to be implemented in a subclass.")

    def __repr__(self):
        return "%s()" % self.__class__.__name__

    def __eq__(self, other):
        if not isinstance(self, other.__class__):
            return NotImplemented
        return True

    def __ne__(self, other):
        are_equal = self.__eq__(other)
        if are_equal is NotImplemented:
            return NotImplemented
        return not are_equal

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))


class ValidationSuccess(ValidationResult):
    @property
    def is_successful(self):
        return True


class ValidationFailure(ValidationResult):
    def __init__(self, invalidity_reason, invalid_parameters):
        self._invalidity_reason = invalidity_reason
        self._invalid_parameters = invalid_parameters

    @property
    def is_successful(self):
        return False

    @property
    def invalidity_reason(self):
        return self._invalidity_reason

    @property
    def invalid_parameters(self):
        return self._invalid_parameters

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.invalidity_reason, self.invalid_parameters)

    def __eq__(self, other):
        if not isinstance(self, other.__class__):
            return NotImplemented
        return self.invalidity_reason == other.invalidity_reason and self.invalid_parameters == other.invalid_parameters



class EventNotificationException(Exception):
    pass


class EventNotification(object):
    STARTED = "started"
    ENDED = "ended"
    SUPPORTED_STATUSES = [STARTED, ENDED]

    def __init__(self, action, status, parameters):
        if status not in self.SUPPORTED_STATUSES:
            raise EventNotificationException(
                "Expected status to be one of %s but got '%s'." % (self.SUPPORTED_STATUSES, status)
            )
        self._action = action
        self._status = status
        self._parameters = parameters

    @property
    def action(self):
        return self._action

    @property
    def status(self):
        return self._status

    @property
    def parameters(self):
        return self._parameters

    def __eq__(self, other):
        return other.action == self.action and other.status == self.status and other.parameters == self.parameters

    def __ne__(self, other):
        return not (self == other)

    def __unicode__(self):
        return "EventNotification(%r, %r, %r)" % (self._action, self._status, self._parameters)



from tala.utils.as_json import convert_to_json, JSONLoggable


class Set(JSONLoggable):
    def __init__(self, content_or_contentclass=None):
        super(Set, self).__init__()
        if content_or_contentclass is None:
            contentclass = None
            initial_content = []
        elif hasattr(content_or_contentclass, '__iter__'):
            contentclass = None
            initial_content = content_or_contentclass
        else:
            contentclass = content_or_contentclass
            initial_content = []
        self.contentclass = contentclass
        self.content = []
        for x in initial_content:
            self.add(x)

    def __unicode__(self):
        return "{" + ", ".join(map(unicode, self.content)) + "}"

    def __str__(self):
        return unicode(self).encode("utf-8")

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.content or self.contentclass or [])

    def as_json(self):
        return {
            "set": list(convert_to_json(object_) for object_ in self.content),
        }

    def __eq__(self, other):
        try:
            return self.is_subset_of(other) and other.is_subset_of(self)
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self.__class__.__name__) + hash(self.content)

    def __ne__(self, other):
        return not (self == other)

    def add(self, element):
        if element not in self.content:
            self._typecheck(element)
            self.content.append(element)

    def remove(self, element):
        self.content.remove(element)

    def remove_if_exists(self, element):
        if element in self.content:
            self.remove(element)

    def is_subset_of(self, other):
        for item in self.content:
            if item not in other.content:
                return False
        return True

    def _typecheck(self, element):
        if self.contentclass:
            if not isinstance(element, self.contentclass):
                raise TypeError(
                    "object " + unicode(element) + " of type " + element.__class__.__name__ + " is not of type " +
                    unicode(self.contentclass)
                )

    def __len__(self):
        return len(self.content)

    def isEmpty(self):
        return len(self) == 0

    def clear(self):
        self.content = []

    def __iter__(self):
        return self.content.__iter__()

    def union(self, other):
        union_set = Set()
        for item in self:
            union_set.add(item)
        for item in other:
            union_set.add(item)
        return union_set

    def extend(self, other):
        for item in other:
            self.add(item)

    def intersection(self, other):
        intersection_set = Set()
        for item in self:
            if item in other:
                intersection_set.add(item)
        return intersection_set



def convert_to_json(object_):
    if object_ is None:
        return None
    if object_ is True or object_ is False:
        return object_
    if isinstance(object_, list):
        return [convert_to_json(element) for element in object_]
    if isinstance(object_, set):
        return {
            "set": [convert_to_json(element) for element in object_]
        }
    if isinstance(object_, dict):
        return {str(key): convert_to_json(value) for key, value in object_.items()}
    if isinstance(object_, JSONLoggable):
        return object_.as_json()
    return unicode(object_)


class JSONLoggable(object):
    @property
    def can_convert_to_json(self):
        return True

    def as_json(self):
        dict_ = {}
        for key, value in self.__dict__.items():
            dict_[key] = convert_to_json(value)
        return dict_



import os
import shutil
import unittest
import tempfile

from tala.ddd.maker import ddd_maker


class TestDddMaker(unittest.TestCase):
    def setUp(self):
        self._temp_dir = tempfile.mkdtemp(prefix="TestDddMaker")
        self._cwd = os.getcwd()

    def tearDown(self):
        os.chdir(self._cwd)
        shutil.rmtree(self._temp_dir)

    def test_custom_target_dir(self):
        self._given_target_dir_specified()
        self._when_make_is_called()
        self._then_everything_is_created()

    def test_default_target_dir(self):
        self._given_default_target_dir()
        self._when_make_is_called()
        self._then_everything_is_created()

    def _then_everything_is_created(self):
        self._base_dir_is_created()
        self._domain_is_created()
        self._ontology_is_created()
        self._service_interface_is_created()
        self._backend_config_is_created()
        self._ddd_config_is_created()
        self._interaction_tests_are_created()
        self._word_list_is_created()

    def _given_target_dir_specified(self):
        self._target_dir = "%s/mock_ddd" % self._temp_dir

    def _given_default_target_dir(self):
        os.chdir(self._temp_dir)
        self._target_dir = "."

    def _when_make_is_called(self):
        ddd_maker.DddMaker("Ddd", "ddd", self._target_dir).make()
        os.chdir(self._target_dir)

    def _base_dir_is_created(self):
        self.assertTrue(os.path.exists("ddd"))

    def _domain_is_created(self):
        self.assertTrue(os.path.exists("ddd/domain.xml"))

    def _ontology_is_created(self):
        self.assertTrue(os.path.exists("ddd/ontology.xml"))

    def _service_interface_is_created(self):
        self.assertTrue(os.path.exists("ddd/service_interface.xml"))

    def _ddd_config_is_created(self):
        self.assertTrue(os.path.exists("ddd/ddd.config.json"))

    def _backend_config_is_created(self):
        self.assertTrue(os.path.exists("backend.config.json"))

    def _interaction_tests_are_created(self):
        self.assertTrue(os.path.exists("ddd/test/interaction_tests_eng.txt"))

    def _word_list_is_created(self):
        self.assertTrue(os.path.exists("ddd/word_list.txt"))



import imp
import inspect
import os

from tala.utils.unicodify import unicodify


class DddLoaderException(Exception):
    pass


class MissingModuleException(Exception):
    pass


class PythonModuleLoader(object):
    def __init__(self, name):
        self._name = name

    def load_py_module_class(self, module_path, ParentClass):
        if os.path.exists(module_path):
            module = self._load_py_module(module_path)
            class_name = self._find_subclass_in_module(ParentClass, module)
            return getattr(module, class_name)
        else:
            raise MissingModuleException("%s missing in DDD %r at %r" % (module_path, self._name, os.getcwd()))

    def _load_py_module(self, module_filename):
        module_name = os.path.splitext(module_filename)[0]
        ddd_module_name = "%s_%s" % (self._name, module_name)
        try:
            return imp.load_source(ddd_module_name, module_filename)
        except IOError:
            raise DddLoaderException("failed to load module %r from %s" % (module_filename, os.getcwd()))

    def _find_subclass_in_module(self, BaseClass, module):
        all_module_classes = inspect.getmembers(module, inspect.isclass)
        classes = self._find_classes_defined_in_module(all_module_classes, module)
        subclasses = self._classes_of_base_class(classes, BaseClass)
        if not subclasses:
            raise DddLoaderException(
                "Could not find class %s in file %r" % (BaseClass.__name__, os.path.abspath(module.__file__))
            )
        if len(subclasses) > 1:
            raise DddLoaderException(
                "Expected one class with base class %r but found %s" % (BaseClass.__name__, unicodify(subclasses))
            )
        subclass = subclasses[0]
        return subclass.__name__

    def _find_classes_defined_in_module(self, classes, module):
        def is_class_defined_in_module(class_, module):
            return inspect.getmodule(class_).__name__ == module.__name__

        return [class_ for name, class_ in classes if is_class_defined_in_module(class_, module)]

    def _classes_of_base_class(self, classes, BaseClass):
        def base_classes_of(class_):
            return inspect.getmro(class_)

        return [class_ for class_ in classes if BaseClass in base_classes_of(class_)]



import copy
import re

from tala.model.error import OntologyError
from tala.model.individual import Individual, NegativeIndividual
from tala.model.lambda_abstraction import LambdaAbstractedPredicateProposition
from tala.model.proposition import PredicateProposition
from tala.model.question import YesNoQuestion, WhQuestion
from tala.model.action import Action
from tala.model.sort import DomainSort, RealSort, IntegerSort, StringSort, BooleanSort, ImageSort, WebviewSort, DateTimeSort, InvalidValueException
from tala.model.image import Image
from tala.model.webview import Webview
from tala.model.date_time import DateTime
from tala.utils.as_json import JSONLoggable


class IndividualExistsException(Exception):
    pass


class SortDoesNotExistException(Exception):
    pass


class ActionDoesNotExistException(Exception):
    pass


class PredicateDoesNotExistException(Exception):
    pass


class InvalidIndividualName(Exception):
    pass


class AmbiguousNamesException(Exception):
    pass


individual_name_mather = re.compile("^[a-zA-Z0-9_]{1,}$")


class DddOntology:
    pass


class Ontology(JSONLoggable):
    DEFAULT_ACTIONS = {"top", "up"}

    def __init__(self, name, sorts, predicates, individuals, actions):
        super(Ontology, self).__init__()
        self._name = name
        self._sorts = self._set_to_dict(sorts)
        self._predicates = self._set_to_dict(predicates)
        self._individuals = individuals
        self._original_individuals = copy.deepcopy(individuals)
        self._actions = self.DEFAULT_ACTIONS.union(actions)
        self._add_default_sorts()
        self._add_builtin_sorts_for_predicates()
        self._validate_individuals()
        self._validate_predicates()
        self._check_predicate_action_integrity()

    def as_json(self):
        json = super(Ontology, self).as_json()
        json["_original_individuals"] = "<skipped>"
        return json

    def reset(self):
        self._individuals = copy.deepcopy(self._original_individuals)

    @property
    def name(self):
        return self._name

    def _set_to_dict(self, set_):
        dict_ = {}
        for item in set_:
            dict_[item.get_name()] = item
        return dict_

    def _validate_individuals(self):
        for name, sort in self._individuals.iteritems():
            self._validate_individual_name(name)
            self._validate_individual_sort(name, sort)

    def is_individual_static(self, name):
        return name in self._original_individuals

    def _validate_individual_name(self, name):
        if not individual_name_mather.match(name):
            raise InvalidIndividualName("invalid individual name %r" % name)

    def _validate_individual_sort(self, name, sort):
        if sort not in self._sorts.values():
            raise OntologyError("individual '%s' has unknown sort '%s' (%s)" % (name, sort.get_name(), self.__dict__))

    def _add_default_sorts(self):
        self._sorts["domain"] = DomainSort()

    def _add_builtin_sorts_for_predicates(self):
        for predicate in self._predicates.values():
            self._add_builtin_sorts_for_predicate(predicate)

    def _add_builtin_sorts_for_predicate(self, predicate):
        sort = predicate.getSort()
        if sort.is_builtin() and sort not in self._sorts:
            self._sorts[sort.get_name()] = sort

    def _validate_predicates(self):
        for predicate in self._predicates.values():
            if predicate.getSort() not in self._sorts.values():
                raise OntologyError(
                    "predicate '%s' has unknown sort '%s' (sorts=%s)" %
                    (predicate.get_name(), predicate.getSort(), self._sorts)
                )

    def _check_predicate_action_integrity(self):
        for action in self._actions:
            if action in self._predicates:
                raise AmbiguousNamesException(
                    "Action '%s' is also a predicate name in %s ontology." % (action, self.name)
                )

    def get_sorts(self):
        return self._sorts

    def get_predicates(self):
        return self._predicates

    def get_actions(self):
        return self._actions

    def get_ddd_specific_actions(self):
        return self._actions.difference(self.DEFAULT_ACTIONS)

    def get_action(self, name):
        if name not in self._actions:
            raise ActionDoesNotExistException("Expected one of the known sorts %s but got '%s'" % (self._actions, name))
        else:
            return Action(name, ontology_name=self.name)

    def get_individuals(self):
        return self._individuals

    def get_individuals_of_sort(self, expected_sort):
        return [individual for individual, sort in self._individuals.iteritems() if sort.get_name() == expected_sort]

    def set_individuals(self, individuals):
        self._individuals = individuals

    def add_individual(self, name, sort_as_string):
        if name in self._individuals.keys():
            raise IndividualExistsException("invididual %r already exists" % name)
        self._validate_name_and_add_individual(name, sort_as_string)

    def _validate_name_and_add_individual(self, name, sort_as_string):
        self._validate_individual_name(name)
        self._individuals[name] = self.get_sort(sort_as_string)

    def ensure_individual_exists(self, name, sort_as_string):
        if name not in self._individuals.keys():
            self._validate_name_and_add_individual(name, sort_as_string)

    def get_name(self):
        return self.name

    def __unicode__(self):
        return self.name

    def __str__(self):
        return unicode(self).encode('utf-8')

    def get_sort(self, name):
        if not self.has_sort(name):
            raise SortDoesNotExistException("unknown sort '%s' in %s" % (name, self))
        else:
            return self._sorts[name]

    def get_predicate(self, name):
        try:
            return self._predicates[name]
        except KeyError:
            raise OntologyError("unknown predicate %r in %s" % (name, self))

    def has_predicate(self, name):
        return name in self._predicates.keys()

    def has_sort(self, name):
        return name in self._sorts.keys()

    def isPredicate(self, predicate):
        return predicate in self._predicates.values()

    def individual_sort(self, value):
        def sort_of(value):
            if self._individual_is_of_enumerated_sort(value):
                return self._get_enumerated_individual_sort(value)
            if self._is_float_value(value):
                return RealSort()
            if self._is_boolean_value(value):
                return BooleanSort()
            if self._is_integer_value(value):
                return IntegerSort()
            if isinstance(value, Image):
                return ImageSort()
            if isinstance(value, Webview):
                return WebviewSort()
            if self._is_string_value(value):
                return StringSort()
            if isinstance(value, DateTime):
                return DateTimeSort()
            raise OntologyError(
                "Expected a value of an individual, or matching one of the builtin sorts, but got '{}'".format(value)
            )

        sort = sort_of(value)
        if not self.predicates_contain_sort(sort.name):
            raise OntologyError(
                "Expected one of the predicate sorts {} in ontology '{}', but got value '{}' of sort '{}'".format(
                    self._predicate_sorts, self.name, value, sort.name
                )
            )
        return sort

    def _individual_is_of_enumerated_sort(self, value):
        return value in self._individuals

    def _get_enumerated_individual_sort(self, value):
        try:
            return self._individuals[value]
        except KeyError:
            raise OntologyError("failed to get sort of unknown individual: " + unicode(value))

    def predicates_contain_sort(self, sort):
        return sort in self._predicate_sorts

    @property
    def _predicate_sorts(self):
        return {predicate.getSort().get_name() for predicate in self._predicates.values()}

    def _is_float_value(self, value):
        return isinstance(value, float)

    def _is_integer_value(self, value):
        return isinstance(value, int)

    def _is_boolean_value(self, value):
        try:
            BooleanSort().normalize_value(value)
        except InvalidValueException:
            return False
        return True

    def _is_string_value(self, value):
        try:
            StringSort().normalize_value(value)
        except InvalidValueException:
            return False
        return True

    def is_action(self, value):
        return value in self._actions

    def create_lambda_abstracted_predicate_proposition(self, predicate):
        if not self.isPredicate(predicate):
            raise OntologyError("predicate %s is not valid in ontology %s" % (unicode(predicate), self.name))
        return LambdaAbstractedPredicateProposition(predicate, self.name)

    def create_action(self, action):
        if action not in self._actions:
            raise OntologyError("Expected one of %s but got %r" % (self._actions, action))
        return Action(action, self.name)

    def create_individual(self, value, sort=None):
        if sort is None:
            sort = self.individual_sort(value)
        value = self._normalize_and_assert_valid_individual_value(value, sort)
        return Individual(self.name, value, sort)

    def _normalize_and_assert_valid_individual_value(self, value, sort):
        if sort.is_dynamic():
            self._assume_value_of_dynamic_sort_is_valid()
            return value
        else:
            return self._normalize_and_assert_valid_individual_value_of_non_dynamic_sort(value, sort)

    def _assume_value_of_dynamic_sort_is_valid(self):
        pass

    def _normalize_and_assert_valid_individual_value_of_non_dynamic_sort(self, value, sort):
        if sort.is_builtin():
            return sort.normalize_value(value)
        else:
            self._assert_valid_individual_value_of_non_dynamic_custom_sort(value)
            return value

    def _assert_valid_individual_value_of_non_dynamic_custom_sort(self, value):
        try:
            self.individual_sort(value)
        except OntologyError as error:
            raise OntologyError("failed to create individual with unknown value: " + unicode(error))

    def create_negative_individual(self, value):
        try:
            sort = self.individual_sort(value)
        except OntologyError:
            raise OntologyError("failed to create negative individual with unknown value: %r" % value)
        return NegativeIndividual(self.name, value, sort)

    def create_yes_no_question(self, predicate_name, individual_value):
        predicate = self.get_predicate(predicate_name)
        individual = self.create_individual(individual_value)
        proposition = self.create_predicate_proposition(predicate, individual)
        return YesNoQuestion(proposition)

    def create_wh_question(self, predicate_name):
        predicate = self.get_predicate(predicate_name)
        lambda_abstracted_prop = self.create_lambda_abstracted_predicate_proposition(predicate)
        return WhQuestion(lambda_abstracted_prop)

    def create_predicate_proposition(self, predicate, individual):
        proposition = PredicateProposition(predicate, individual)
        return proposition



import os

from pathlib import Path


class UTF8FileWriter(object):
    def __init__(self, path):
        super(UTF8FileWriter, self).__init__()
        self._path = path if isinstance(path, Path) else Path(path)

    def create_directories(self):
        if not self._path.parent.exists():
            os.makedirs(str(self._path.parent))

    def write(self, data):
        with self._path.open("w", encoding="utf-8") as file_:
            file_.write(unicode(data))



from jinja2 import Template

from tala.ddd.services.service_interface import ServiceInterface, ServiceActionInterface, ServiceParameter, DeviceModuleTarget, ActionFailureReason, ServiceQueryInterface, FrontendTarget, PlayAudioActionInterface, AudioURLServiceParameter, ServiceEntityRecognizerInterface, ServiceValidatorInterface
from tala.model.device import DeviceAction


class ServiceInterfaceFromDevice(object):
    @classmethod
    def from_device_handler(cls, device_handler):
        play_audio_actions = list(cls._convert_play_audio_actions(device_handler))
        custom_actions = list(cls._convert_actions(device_handler))
        actions = play_audio_actions + custom_actions
        queries = list(cls._convert_queries(device_handler))
        validators = list(cls._convert_validators(device_handler))
        entity_recognizers = list(cls._convert_entity_recognizers(device_handler))
        service_interface = ServiceInterface(actions, queries, entity_recognizers, validators)
        return service_interface

    @classmethod
    def _convert_play_audio_actions(cls, device_handler):
        actions = device_handler.get_actions()
        for name, action in actions.iteritems():
            if hasattr(action, "TYPE") and action.TYPE == DeviceAction.PLAY_AUDIO:
                audio_url_predicate = action.AUDIO_URL_PARAMETER
                audio_url_parameter = AudioURLServiceParameter(audio_url_predicate)
                all_parameters = list(cls.convert_parameters(action.parameters))
                parameters = [parameter for parameter in all_parameters if parameter.name != audio_url_predicate]
                yield PlayAudioActionInterface(name, FrontendTarget(), parameters, audio_url_parameter)

    @classmethod
    def _convert_actions(cls, device_handler):
        actions = device_handler.get_actions()
        for name, action in actions.iteritems():
            if not hasattr(action, "TYPE") or action.TYPE != DeviceAction.PLAY_AUDIO:
                target = DeviceModuleTarget(device_handler.get_name())
                parameters = list(cls.convert_parameters(action.parameters))
                device_failure_reasons = action.FAILURE_REASONS if hasattr(action, "FAILURE_REASONS") else []
                failure_reasons = list(cls._convert_failure_reasons(device_failure_reasons))
                yield ServiceActionInterface(name, target, parameters, failure_reasons)

    @classmethod
    def _convert_queries(cls, device_handler):
        queries = device_handler.get_queries()
        device_name = device_handler.get_name()
        for name, query in queries.iteritems():
            parameters = list(cls.convert_parameters(query.parameters))
            yield ServiceQueryInterface(name, DeviceModuleTarget(device_name), parameters)

    @classmethod
    def _convert_validators(cls, device_handler):
        validities = device_handler.get_validities()
        device_name = device_handler.get_name()
        for name, validity in validities.iteritems():
            parameters = list(cls.convert_parameters(validity.parameters))
            yield ServiceValidatorInterface(name, DeviceModuleTarget(device_name), parameters)

    @classmethod
    def _convert_entity_recognizers(cls, device_handler):
        entity_recognizers = device_handler.get_entity_recognizers()
        device_name = device_handler.get_name()
        for name, entity_recognizer in entity_recognizers.iteritems():
            yield ServiceEntityRecognizerInterface(name, DeviceModuleTarget(device_name))

    @classmethod
    def convert_parameters(cls, device_parameters):
        for parameter in device_parameters:
            name = parameter["name"]
            format = parameter["field"]
            is_optional = parameter["default_value"] is not None
            yield ServiceParameter(name, format, is_optional)

    @classmethod
    def _convert_failure_reasons(cls, device_failure_reasons):
        for reason in device_failure_reasons:
            yield ActionFailureReason(reason)

    @staticmethod
    def to_xml(service_interface):
        template = Template(
            """{%- macro parameters(parameters) %}
{%- if parameters -%}
    <parameters>
{%- for parameter in parameters %}
      <parameter predicate="{{ parameter.name }}" format="{{ parameter.format }}" optional="{% if parameter.is_optional %}true{% else %}false{% endif %}"/>
{%- endfor %}
    </parameters>
{%- else -%}
    <parameters/>
{%- endif -%}
{%- endmacro -%}
{%- macro failure_reasons(failure_reasons) %}
{%- if failure_reasons -%}
    <failure_reasons>
{%- for failure_reason in failure_reasons %}
      <failure_reason name="{{ failure_reason.name }}"/>
{%- endfor %}
    </failure_reasons>
{%- else -%}
    <failure_reasons/>
{%- endif %}
{%- endmacro %}
{%- macro target(target) -%}
    <target>
{%- if target.is_device_module %}
      <device_module device="{{ target.device }}"/>
{%- else %}
      <frontend/>
{%- endif %}
    </target>
{%- endmacro -%}
<?xml version="1.0" encoding="utf-8"?>
<service_interface>
{%- for action in service_interface.actions %}
  <action name="{{ action.name }}">
    {{ parameters(action.parameters) }}
    {{ failure_reasons(action.failure_reasons) }}
    {{ target(action.target) }}
  </action>
{%- endfor %}
{%- for query in service_interface.queries %}
  <query name="{{ query.name }}">
    {{ parameters(query.parameters) }}
    {{ target(query.target) }}
  </query>
{%- endfor %}
{%- for validator in service_interface.validators %}
  <validator name="{{ validator.name }}">
    {{ parameters(validator.parameters) }}
    {{ target(validator.target) }}
  </validator>
{%- endfor %}
{%- for entity_recognizer in service_interface.entity_recognizers %}
  <entity_recognizer name="{{ entity_recognizer.name }}">
    {{ target(entity_recognizer.target) }}
  </entity_recognizer>
{%- endfor %}
</service_interface>
"""
        )
        written_interface = template.render(service_interface=service_interface)
        return written_interface



class AudioDirective(object):
    PLAY = "PLAY"
    STOP = "STOP"

    def __init__(self, type):
        self._type = type

    @property
    def type(self):
        return self._type


class AudioPlayDirective(AudioDirective):
    def __init__(self, url, offset=0):
        super(AudioPlayDirective, self).__init__(AudioDirective.PLAY)
        self._url = url
        self._offset = offset

    @property
    def url(self):
        return self._url

    @property
    def offset(self):
        return self._offset

    def __eq__(self, other):
        return other.url == self.url and other.offset == self.offset

    def __repr__(self):
        return "AudioPlayDirective(%r, %r)" % (self.url, self.offset)


class AudioStopDirective(AudioDirective):
    def __init__(self):
        super(AudioStopDirective, self).__init__(AudioDirective.STOP)

    def __repr__(self):
        return "AudioStopDirective()"



import json
import os
import unittest
import tempfile

from mock import patch

from tala.config import Config, DddConfig, RasaConfig, BackendConfig, UnexpectedConfigEntriesException, BackendConfigNotFoundException, DddConfigNotFoundException, RasaConfigNotFoundException, DEFAULT_INACTIVE_SECONDS_ALLOWED


class ConfigTester(object):
    def setup_configs(self):
        self.configs = {}
        self._working_dir = os.getcwd()
        temp_dir = tempfile.mkdtemp(prefix="TestConfig")
        os.chdir(temp_dir)

    def teardown_configs(self):
        os.chdir(self._working_dir)

    def given_config_file_exists(self, *args, **kwargs):
        self._create_config_file(*args, **kwargs)

    def _create_config_file(self, ConfigClass, config, name=None):
        if name is None:
            name = ConfigClass.default_name()
        self._write_config(name, config)

    def _write_config(self, name, config):
        with open(name, 'w') as f:
            json.dump(config, f)

    def given_created_config(self, ConfigClass, path=None):
        self._config = ConfigClass(path)

    def given_default_config(self, mock_default_config, value):
        mock_default_config.return_value = value

    def given_default_name(self, mock_default_name, value):
        mock_default_name.return_value = value

    def when_trying_to_read(self):
        try:
            self.when_reading()
        except UnexpectedConfigEntriesException:
            pass

    def when_reading_then_exception_is_raised_with_message_that_matches_regex(self, exception_class, expected_regex):
        with self.assertRaisesRegexp(exception_class, expected_regex):
            self.when_reading()

    def when_reading(self):
        self._read_config = self._config.read()

    def then_read_config_was(self, config):
        self.assertDictEqual(self._read_config, config)

    def then_config_is(self, filename, expected_config):
        with open(filename) as f:
            actual_config = json.load(f)
        self.assertDictEqual(actual_config, expected_config)

    def then_wrote(self, filename, expected_config):
        with open(filename) as f:
            actual_config = json.load(f)
        self.assertDictEqual(expected_config, actual_config)


class TestConfig(unittest.TestCase, ConfigTester):
    def setUp(self):
        self.setup_configs()

    def tearDown(self):
        self.teardown_configs()

    @patch.object(Config, "default_config")
    @patch.object(Config, "default_name")
    def test_read(self, mock_default_name, mock_default_config):
        self.given_default_config(mock_default_config, {"mock_key": "mock_value"})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(Config, {"mock_key": "mock_value"})
        self.given_created_config(Config)
        self.when_reading()
        self.then_read_config_was({"mock_key": "mock_value"})

    @patch.object(Config, "default_config")
    @patch.object(Config, "default_name")
    def test_read_with_unexpected_and_missing_entry_raises_exception(self, mock_default_name, mock_default_config):
        self.given_default_config(mock_default_config, {"mock_key": "mock_value"})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(Config, {"mock_unexpected_key": ""})
        self.given_created_config(Config)
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            UnexpectedConfigEntriesException,
            "Parameter mock_unexpected_key is unexpected while mock_key is missing from.*"
            "The config was updated and the previous config was backed up in 'mock\.config\.json\.backup'\."
        )

    @patch.object(Config, "default_config")
    @patch.object(Config, "default_name")
    def test_read_with_unexpected_entry_raises_exception(self, mock_default_name, mock_default_config):
        self.given_default_config(mock_default_config, {"mock_key": "mock_value"})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(Config, {"mock_key": "mock_value", "mock_unexpected_key": ""})
        self.given_created_config(Config)
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            UnexpectedConfigEntriesException, "Parameter mock_unexpected_key is unexpected in.*"
            "The config was updated and the previous config was backed up in 'mock\.config\.json\.backup'\."
        )

    @patch.object(Config, "default_config")
    @patch.object(Config, "default_name")
    def test_read_with_missing_entries_raises_exception(self, mock_default_name, mock_default_config):
        self.given_default_config(mock_default_config, {"mock_key_1": "mock_value", "mock_key_2": "mock_value"})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(Config, {})
        self.given_created_config(Config)
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            UnexpectedConfigEntriesException, "Parameter mock_key_1 and mock_key_2 is missing from.*"
            "The config was updated and the previous config was backed up in 'mock\.config\.json\.backup'\."
        )

    @patch.object(Config, "default_config")
    @patch.object(Config, "default_name")
    def test_read_with_missing_entries_saves_old_config(self, mock_default_name, mock_default_config):
        self.given_default_config(mock_default_config, {"mock_key": "mock_value"})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(Config, {"mock_unexpected_key": "mock_unexpected_value"})
        self.given_created_config(Config)
        self.when_trying_to_read()
        self.then_config_is(self._config.back_up_name(), {"mock_unexpected_key": "mock_unexpected_value"})

    @patch.object(Config, "default_config")
    @patch.object(Config, "default_name")
    def test_read_with_missing_entries_writes_updated_config(self, mock_default_name, mock_default_config):
        self.given_default_config(
            mock_default_config, {
                "mock_key_1": "mock_default_value_1",
                "mock_key_2": "mock_default_value_2"
            }
        )
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(
            Config, {
                "mock_unexpected_key": "mock_unexpected_value",
                "mock_key_2": "mock_updated_value_2"
            }
        )
        self.given_created_config(Config)
        self.when_trying_to_read()
        self.then_config_is(
            Config.default_name(), {
                "mock_key_1": "mock_default_value_1",
                "mock_key_2": "mock_updated_value_2"
            }
        )

    @patch.object(Config, "default_config")
    @patch.object(Config, "default_name")
    def test_write_default_config(self, mock_default_name, mock_default_config):
        self.given_default_config(mock_default_config, {"mock_key": "mock_value"})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.when_writing_default_config(Config)
        self.then_wrote("mock.config.json", {"mock_key": "mock_value"})

    def when_writing_default_config(self, Config):
        Config.write_default_config()


class TestInheritance(unittest.TestCase, ConfigTester):
    def setUp(self):
        self.setup_configs()

    def tearDown(self):
        self.teardown_configs()

    @patch.object(Config, "default_config")
    def test_override_value_from_parent(self, mock_default_config):
        self.given_default_config(mock_default_config, {"key_to_override": "mock_default_value", "overrides": None})
        self.given_config_file_exists(
            Config, {
                "key_to_override": "parent_value",
                "overrides": None
            }, name="parent.config.json"
        )
        self.given_config_file_exists(
            Config, {
                "overrides": "parent.config.json",
                "key_to_override": "child_value"
            }, name="child.config.json"
        )
        self.given_created_config(Config, "child.config.json")
        self.when_reading()
        self.then_read_config_was({"key_to_override": "child_value", "overrides": "parent.config.json"})

    @patch.object(Config, "default_config")
    def test_keep_value_from_parent(self, mock_default_config):
        self.given_default_config(mock_default_config, {"key_to_override": "mock_default_value", "overrides": None})
        self.given_config_file_exists(
            Config, {
                "key_to_override": "parent_value",
                "overrides": None
            }, name="parent.config.json"
        )
        self.given_config_file_exists(Config, {"overrides": "parent.config.json"}, name="child.config.json")
        self.given_created_config(Config, "child.config.json")
        self.when_reading()
        self.then_read_config_was({"key_to_override": "parent_value", "overrides": "parent.config.json"})

    def test_read_with_missing_parent_raises_exception(self):
        self.given_config_file_exists(
            BackendConfig, {"overrides": "non_existing_parent.config.json"}, name="child.config.json"
        )
        self.given_created_config(BackendConfig, "child.config.json")
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            BackendConfigNotFoundException,
            "Expected backend config '.*non_existing_parent\.config\.json' to exist but it was not found\."
        )


class TestDddConfig(unittest.TestCase, ConfigTester):
    def setUp(self):
        self.setup_configs()

    def tearDown(self):
        self.teardown_configs()

    def test_default_name(self):
        self.assertEqual(DddConfig.default_name(), "ddd.config.json")

    def test_default_config(self):
        self.assertEqual(
            DddConfig.default_config(), {
                "use_rgl": True,
                "use_third_party_parser": False,
                "device_module": None,
                "word_list": "word_list.txt",
                "enable_rasa_nlu": False,
                "overrides": None,
            }
        )

    def test_read_with_missing_config_file_raises_exception(self):
        self.given_created_config(DddConfig, "non_existing_config.json")
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            DddConfigNotFoundException,
            "Expected DDD config '.*non_existing_config\.json' to exist but it was not found\."
        )

    def given_default_ddd_config_with_rasa_enabled_and_rgl_disabled(self):
        self._json_config = DddConfig.default_config()
        self._json_config["enable_rasa_nlu"] = True
        self._json_config["use_rgl"] = False

    def given_config_file_written(self):
        self._create_config_file(DddConfig, self._json_config)


class TestRasaConfig(unittest.TestCase, ConfigTester):
    def setUp(self):
        self.setup_configs()

    def tearDown(self):
        self.teardown_configs()

    def test_default_name(self):
        self.assertEqual(RasaConfig.default_name(), "rasa.config.json")

    def test_default_config(self):
        self.assertEqual(
            RasaConfig.default_config(), {
                "enable_duckling": False,
                "duckling_url": "http://127.0.0.1:10090",
            }
        )

    def test_read_with_missing_config_file_raises_exception(self):
        self.given_created_config(RasaConfig, "non_existing_config.json")
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            RasaConfigNotFoundException,
            "Expected RASA config '.*non_existing_config\.json' to exist but it was not found\."
        )


class TestBackendConfig(unittest.TestCase, ConfigTester):
    def setUp(self):
        self.setup_configs()

    def tearDown(self):
        self.teardown_configs()

    def test_default_name(self):
        self.assertEqual(BackendConfig.default_name(), "backend.config.json")

    def test_default_config(self):
        self.assertEqual(
            BackendConfig.default_config(), {
                "asr": "none",
                "supported_languages": ["eng"],
                "ddds": [""],
                "active_ddd": "",
                "use_recognition_profile": False,
                "repeat_questions": True,
                "use_word_list_correction": False,
                "overrides": None,
                "rerank_amount": 0.2,
                "inactive_seconds_allowed": DEFAULT_INACTIVE_SECONDS_ALLOWED,
            }
        )

    def test_default_config_with_ddd_name(self):
        self.assertEqual(
            BackendConfig.default_config(ddd_name="my-ddd"), {
                "asr": "none",
                "supported_languages": ["eng"],
                "ddds": ["my-ddd"],
                "active_ddd": "my-ddd",
                "use_recognition_profile": False,
                "repeat_questions": True,
                "use_word_list_correction": False,
                "overrides": None,
                "rerank_amount": 0.2,
                "inactive_seconds_allowed": DEFAULT_INACTIVE_SECONDS_ALLOWED,
            }
        )

    def test_read_with_missing_config_file_raises_exception(self):
        self.given_created_config(BackendConfig, "non_existing_config.json")
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            BackendConfigNotFoundException,
            "Expected backend config '.*non_existing_config\.json' to exist but it was not found\."
        )

    def test_read_with_android_asr_and_missing_use_word_list_correction_sets_it_to_true(self):
        self.given_config_file_exists(
            BackendConfig, {
                "asr": "android",
                "supported_languages": ["eng"],
                "ddds": ["my-ddd"],
                "active_ddd": "my-ddd",
                "use_recognition_profile": False,
                "repeat_questions": True,
                "overrides": None,
                "inactive_seconds_allowed": DEFAULT_INACTIVE_SECONDS_ALLOWED,
            }
        )
        self.given_created_config(BackendConfig)
        self.when_trying_to_read()
        self.then_config_is(
            BackendConfig.default_name(), {
                "asr": "android",
                "supported_languages": ["eng"],
                "ddds": ["my-ddd"],
                "active_ddd": "my-ddd",
                "use_recognition_profile": False,
                "repeat_questions": True,
                "use_word_list_correction": True,
                "overrides": None,
                "rerank_amount": 0.2,
                "inactive_seconds_allowed": DEFAULT_INACTIVE_SECONDS_ALLOWED,
            }
        )

    def test_read_with_android_asr_and_something_missing_keeps_use_word_list_correction_if_false(self):
        self.given_config_file_exists(
            BackendConfig, {
                "asr": "android",
                "supported_languages": ["eng"],
                "ddds": ["my-ddd"],
                "active_ddd": "my-ddd",
                "use_recognition_profile": False,
                "use_word_list_correction": False,
                "overrides": None,
                "inactive_seconds_allowed": DEFAULT_INACTIVE_SECONDS_ALLOWED,
            }
        )
        self.given_created_config(BackendConfig)
        self.when_trying_to_read()
        self.then_config_is(
            BackendConfig.default_name(), {
                "asr": "android",
                "supported_languages": ["eng"],
                "ddds": ["my-ddd"],
                "active_ddd": "my-ddd",
                "use_recognition_profile": False,
                "repeat_questions": True,
                "use_word_list_correction": False,
                "overrides": None,
                "rerank_amount": 0.2,
                "inactive_seconds_allowed": DEFAULT_INACTIVE_SECONDS_ALLOWED,
            }
        )

    def test_read_with_android_asr_and_something_missing_keeps_use_word_list_correction_if_true(self):
        self.given_config_file_exists(
            BackendConfig, {
                "asr": "android",
                "supported_languages": ["eng"],
                "ddds": ["my-ddd"],
                "active_ddd": "my-ddd",
                "use_recognition_profile": False,
                "use_word_list_correction": True,
                "overrides": None,
                "inactive_seconds_allowed": DEFAULT_INACTIVE_SECONDS_ALLOWED,
            }
        )
        self.given_created_config(BackendConfig)
        self.when_trying_to_read()
        self.then_config_is(
            BackendConfig.default_name(), {
                "asr": "android",
                "supported_languages": ["eng"],
                "ddds": ["my-ddd"],
                "active_ddd": "my-ddd",
                "use_recognition_profile": False,
                "repeat_questions": True,
                "use_word_list_correction": True,
                "overrides": None,
                "rerank_amount": 0.2,
                "inactive_seconds_allowed": DEFAULT_INACTIVE_SECONDS_ALLOWED,
            }
        )

    def test_write_default_config_with_ddd(self):
        self.when_writing_default_config_with_ddd_name("my-ddd")
        self.then_wrote(BackendConfig.default_name(), BackendConfig.default_config(ddd_name="my-ddd"))

    def when_writing_default_config_with_ddd_name(self, ddd_name):
        BackendConfig.write_default_config(ddd_name=ddd_name)
